#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project : riboParser
# @Script  : Bam2Wig.py


import pysam
import os
import sys
import numpy as np
import pandas as pd
import polars as pl
from collections import OrderedDict


class Bam2Wig(object):

    def __init__(self, args):
        """
        arguments for these:
        bam file
        psite file
        multiple mapping
        bed file for density
        rpm file

        """
        self.bam = args.bam
        self.sample_suffix = os.path.splitext(args.bam)[1]
        self.read_format = None
        self.pysam_input = None

        self.psite = args.psite
        self.offset = {}

        self.tag_num = args.times
        self.secondary = args.secondary
        self.supplementary = args.supplementary
        self.format = args.format

        self.chrom = OrderedDict()
        self.bed_minus = {}
        self.bed_plus = {}
        self.bed_minus_df = None
        self.bed_plus_df = None

        self.norm = args.norm
        self.merge = args.merge
        self.output = args.output

    def read_offset(self):
        """
        import the offset file to locate the p-site of reads from bam file.

        1. import the offset file
        length psite
        27  12
        28  12
        29  13
        30  13
        ...

        2. make the offset dict for reads trim

        """
        offset = pd.read_csv(self.psite, sep='\t', header=0, names=None)
        offset.dropna(axis=0, inplace=True)
        offset["p_site"] = offset["p_site"].astype(int)
        offset["p_site"] = offset["p_site"] - 1
        self.offset = offset[["length", "p_site"]].groupby(offset["length"])["p_site"].apply(list).T.to_dict()
        print(self.offset, flush=True)

        # check the data type.
        if "third" in offset["ribo"].unique():
            self.profile = "trisome"
        elif "second" in offset["ribo"].unique():
            self.profile = "disome"
        elif "first" in offset["ribo"].unique():
            self.profile = "monosome"
        else:
            print("Unknown type of sequence profile, please re-check the offset file!", flush=True)
            sys.exit()

    def create_chroms_dict(self):
        """
        create a dict to save the chromosome and reads number
        """
        # self.chrom = dict(zip(self.pysam_input.header.references, self.pysam_input.header.lengths))
        for chrom in self.pysam_input.header.references:
            self.chrom[chrom] = 0
            self.bed_minus[chrom] = {}
            self.bed_plus[chrom] = {}

    def set_tag_num(self):
        # set the tag_num to discard the reads aligned too many loci
        if self.tag_num == 0:
            self.tag_num = len(self.mrna_dict)
        else:
            pass

    def convert_rpf_density(self):
        """
        import the bam file and convert the reads to density

        1. import the file
        2. remove the reads out of specified range
        3. remove the reads aligned to multiple loci
        3. retrieve the psite
        4. merge the psite density to chrom_dict

        """
        
        now_rows = 0
        for read in self.pysam_input.fetch(until_eof=True):

            # processing
            now_rows += 1
            if now_rows % 100000 == 0:
                print("now rows:" + str(now_rows), flush=True)
            else:
                pass

            # discard the unmapped reads
            if read.is_unmapped:
                continue

            # discard the secondary mapping
            if  read.is_secondary == self.secondary == True:
                continue
            
            # discard the supplementary mapping
            if  read.is_supplementary == self.secondary == True:
                continue

            # filter the unique reads
            if read.has_tag("NH") and read.get_tag("NH") > self.tag_num:
                continue
            
            # split positive and negative strand
            chrom = read.reference_name
            length = read.infer_read_length()
            
            # split positive and negative strand
            if read.is_reverse:
                try:
                    # start = read.aligned_pairs[self.offset[length][0]][1]
                    start = read.positions[-self.offset[length][0]]
                except KeyError:
                    continue

                self.chrom[chrom] += 1

                try:
                    self.bed_minus[chrom][start] += 1
                except KeyError:
                    self.bed_minus[chrom][start] = 1

            else:
                try:
                    # start = read.aligned_pairs[self.offset[length][0]][1]
                    start = read.positions[self.offset[length][0]]
                except KeyError:
                    continue

                self.chrom[chrom] += 1

                try:
                    self.bed_plus[chrom][start] += 1
                except KeyError:
                    self.bed_plus[chrom][start] = 1

        print("now rows:" + str(now_rows), flush=True)

    def import_bam(self):
        """
        check the file format
        import the alignment file with bam or sam file

        """
        if self.sample_suffix.lower() == ".bam":
            print("import file: {bam}.\n".format(bam=self.bam), flush=True)
            self.read_format = 'rb'
        elif self.sample_suffix.lower() == ".sam":
            print("import file: {bam}.\n".format(bam=self.bam), flush=True)
            self.read_format = 'r'
        else:
            print("Unknown file format, please input the correct bam or sam file.", flush=True)
            sys.exit()

        self.pysam_input = pysam.AlignmentFile(self.bam, self.read_format)
        
        self.create_chroms_dict()

        self.convert_rpf_density()
        
        self.pysam_input.close()
        

    def convert_dict_to_dataframe(self):
        """
        1. split minus and plus strand
        2. for each chrom, convert the dict to dataframe
        3. merge all chrom dataframe

        """

        minus_list = []
        for chrom, density in self.bed_minus.items():
            tmp_df = pd.DataFrame([density]).T.reset_index()
            tmp_df.columns = ["start", "density"]
            tmp_df.insert(loc=0, column="chr", value=chrom)

            minus_list.append(tmp_df)
        self.bed_minus_df = pd.concat(minus_list, axis=0)
        self.bed_minus_df.loc[:, "density"] = -self.bed_minus_df["density"]

        plus_list = []
        for chrom, density in self.bed_plus.items():
            tmp_df = pd.DataFrame([density]).T.reset_index()
            tmp_df.columns = ["start", "density"]
            tmp_df.insert(0, "chr", chrom)

            plus_list.append(tmp_df)
        self.bed_plus_df = pd.concat(plus_list, axis=0)

    def norm_rpm(self):
        """
        1. summary the bam reads count
        2. convert the density to rpm
        rpm = ( density * 1000000 ) / total reads
        """
        if self.norm:
            total_reads = pd.DataFrame([self.chrom]).sum()

            self.bed_minus_df.loc[:, "density"] = self.bed_minus_df["density"] * 1e6 / total_reads[0]
            self.bed_plus_df.loc[:, "density"] = self.bed_plus_df["density"] * 1e6 / total_reads[0]

            self.bed_minus_df.loc[:, "density"] = self.bed_minus_df["density"].round(3)
            self.bed_plus_df.loc[:, "density"] = self.bed_plus_df["density"].round(3)
        else:
            print('Skip!', flush=True)

    def output_bed(self):
        """
        1. drop all INF and NA values
        2. convert the ribosome density to bedgraph / wig format
        3. merge plus / minus strand density
        4. output the density file
        """
        self.bed_minus_df.replace([np.inf, -np.inf], np.nan, inplace=True)
        self.bed_plus_df.replace([np.inf, -np.inf], np.nan, inplace=True)
        
        self.bed_minus_df.dropna(inplace=True)
        self.bed_plus_df.dropna(inplace=True)

        if self.merge:
            bed_df = pd.concat([self.bed_plus_df, self.bed_minus_df], axis=0)
            bed_df.reset_index(drop=True).sort_values(["chr", "start"])

            if self.format == "bedgraph":
                bed_df.insert(loc=2, column="end", value=bed_df["start"] + 1)
                bed_df[["start", "end"]] = bed_df[["start", "end"]].astype(int)

                bed_pl = pl.DataFrame(bed_df)
                bed_pl.write_csv(self.output + ".bedgraph", separator = '\t', has_header = False)
            else:
                bed_df["start"] = bed_df["start"].astype(int)
                bed_pl = pl.DataFrame(bed_df)
                bed_pl.write_csv(self.output + ".wig", separator = '\t', has_header = False)

        else:
            if self.format == 'bedgraph':
                self.bed_minus_df.insert(loc=2, column="end", value=self.bed_minus_df["start"] + 1)
                self.bed_plus_df.insert(loc=2, column="end", value=self.bed_plus_df["start"] + 1)

                self.bed_minus_df[["start", "end"]] = self.bed_minus_df[["start", "end"]].astype(int)
                self.bed_plus_df[["start", "end"]] = self.bed_plus_df[["start", "end"]].astype(int)

                bed_minus_pl = pl.DataFrame(self.bed_minus_df)
                bed_plus_pl = pl.DataFrame(self.bed_plus_df)

                bed_minus_pl.write_csv(self.output + "_minus.bedgraph", separator = '\t', has_header = False)
                bed_plus_pl.write_csv(self.output + "_plus.bedgraph", separator = '\t', has_header = False)
            else:
                self.bed_minus_df["start"] = self.bed_minus_df["start"].astype(int)
                self.bed_plus_df["start"] = self.bed_plus_df["start"].astype(int)
            
                bed_minus_pl = pl.DataFrame(self.bed_minus_df)
                bed_plus_pl = pl.DataFrame(self.bed_plus_df)

                bed_minus_pl.write_csv(self.output + "_minus.wig", separator = '\t', has_header = False)
                bed_plus_pl.write_csv(self.output + "_plus.wig", separator = '\t', has_header = False)




#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project : riboParser
# @Script  : Occupancy.py


import sys
import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from . import RPFs


class Occupancy(object):

    def __init__(self, args):
        # input and output
        self.gene = args.list
        self.rpf = args.rpf
        self.output = args.output
        self.output_all = args.all

        # filter the stable codon coverage range
        self.norm = args.normal
        self.site = args.site
        self.frame = args.frame
        self.rpf_num = args.min
        self.tis = args.tis
        self.tts = args.tts

        # titles
        self.sample_num = 0
        self.sample_name = None
        self.dst_title = None
        self.occ_title = None
        self.rel_title = None

        # for the high expression gene filter
        self.merged_rpf = None
        self.raw_rpf = None
        self.high_gene = None
        self.high_rpf = None
        self.gene_rpf_sum = None
        self.total_rpf_num = None

        # table for codon occupancy
        self.scale = args.scale
        self.total_density = None
        self.occupancy = None

        self.stop = args.stop
        self.codon_dict, self.codon_table = RPFs.codon_table()
        
        if self.stop:
            del self.codon_dict['TAA']
            del self.codon_dict['TAG']
            del self.codon_dict['TGA']
            
            self.codon_table = self.codon_table[self.codon_table['Abbr'] != 'Stop']

    def import_rpf(self):
        # raw_rpf, sample_name, sample_num, merged_rpf, total_rpf_num, gene_rpf_sum, high_gene, high_rpf
        rpf_results = RPFs.import_rpf(rpf_file=self.rpf,
                                      sites=self.site,
                                      frame=self.frame,
                                      sample_num=None,
                                      sample_name=None,
                                      tis=self.tis,
                                      tts=self.tts,
                                      gene=self.gene,
                                      rpf_num=self.rpf_num)
        # self.raw_rpf = rpf_results[0]
        self.sample_name = rpf_results[1]
        self.sample_num = rpf_results[2]
        self.merged_rpf = rpf_results[3]
        # rpf_results[3] = ''
        self.total_rpf_num = rpf_results[4]
        # self.gene_rpf_sum = rpf_results[5].copy()
        self.high_gene = rpf_results[6].copy()
        # self.high_rpf = rpf_results[7].copy()
        del rpf_results

        self.merged_rpf = self.merged_rpf[self.merged_rpf['codon'].isin(self.codon_dict.keys())]

        if self.norm:
            self.merged_rpf[self.sample_name] = self.merged_rpf[self.sample_name] * 1e6 / self.total_rpf_num
            # self.gene_rpf_sum = self.high_rpf.loc[:, ["name"] + self.sample_name].groupby("name")[self.sample_name].sum()

        self.high_rpf = self.merged_rpf.loc[self.merged_rpf['name'].isin(self.high_gene)]
        del self.merged_rpf

        # self.high_rpf.loc[:, self.sample_name] = self.high_rpf[self.sample_name] * 1e6 / self.total_rpf_num

    @staticmethod
    def scale_method(scale, occupancy):
        if scale == 'minmax':
            # relative_occupancy = (occupancy - occupancy.min()) / (occupancy.max() - occupancy.min())
            relative_occupancy = occupancy / occupancy.max()
        elif scale == 'zscore':
            relative_occupancy = (occupancy - occupancy.mean()) / occupancy.std()
        else:
            print('Unknown scale method.', flush=True)
            sys.exit()
            
        return relative_occupancy
    
    def codon_occupancy(self):
        # calc the average density of each gene
        # ave_rpf = self.high_rpf.replace(0, np.nan).groupby('name')[self.sample_name].mean()
        ave_rpf = self.high_rpf.groupby('name')[self.sample_name].mean()

        # calculate the average density of each codon
        high_rpf_group = self.high_rpf.groupby('name')
        now_num = 0
        ave_dst_list = []
        for gene, rpfs in high_rpf_group:
            # counter
            now_num += 1
            if now_num % 500 == 0:
                print('Row: {rows}, {gene}.'.format(rows=now_num, gene=gene), flush=True)

            # prepare the gene df
            temp_rpfs = rpfs.loc[:, self.sample_name].copy()
            # temp_ave_dst = np.divide(temp_rpfs, ave_rpf.loc[gene].to_list(), where=ave_rpf.loc[gene] != 0)
            # where is discarded in the new version of numpy
            temp_ave_dst = np.divide(temp_rpfs, ave_rpf.loc[gene].to_list())
            ave_dst_list.append(temp_ave_dst)

        print('Row: {rows}, {gene}.'.format(rows=now_num, gene=gene), flush=True)
        
        # merge average density
        ave_dst = pd.concat(ave_dst_list, axis=0)
        ave_dst_title = [i + '_ave_dst' for i in self.sample_name]
        ave_dst.columns = ave_dst_title
        high_rpf_ave_dst = pd.concat([self.high_rpf, ave_dst], axis=1)
        codon_density = pd.pivot_table(high_rpf_ave_dst,
                                       index=['name', 'codon'],
                                       values=self.sample_name + ave_dst_title,
                                       aggfunc=np.sum)
        if self.output_all:
            high_rpf_ave_dst.to_csv(self.output + '_rpf_density.txt', sep='\t', index=False)

        codon_density.to_csv(self.output + '_codon_density.txt', sep='\t', index=True)

        # calculate the total density and occupancy (average density)
        self.total_density = high_rpf_ave_dst.replace(0, np.nan).groupby('codon')[ave_dst_title].sum()
        self.occupancy = high_rpf_ave_dst.replace(0, np.nan).groupby('codon')[ave_dst_title].mean()

        # merge the codon count
        for codon in self.codon_dict.keys():
            try:
                self.codon_dict[codon].extend(self.total_density.loc[codon].values.tolist())
                self.codon_dict[codon].extend(self.occupancy.loc[codon].values.tolist())
            except KeyError:
                self.codon_dict[codon].extend([np.nan] * self.sample_num)
                self.codon_dict[codon].extend([np.nan] * self.sample_num)

        self.codon_table = pd.DataFrame(self.codon_dict).T
        self.dst_title = [i + '_density' for i in self.sample_name]
        self.occ_title = [i + '_absolute_occupancy' for i in self.sample_name]
        self.codon_table.columns = ['AA', 'Abbr'] + self.dst_title + self.occ_title

        # z-score standard
        # rel_occ = (self.codon_table[self.occ_title] - self.codon_table[self.occ_title].mean()) / self.codon_table[self.occ_title].std()
        rel_occ = self.scale_method(self.scale, self.codon_table[self.occ_title])
        self.rel_title = [i + '_relative_occupancy' for i in self.sample_name]
        rel_occ.columns = self.rel_title
        self.codon_table = pd.concat([self.codon_table, rel_occ], axis=1)
        self.codon_table.sort_values('Abbr', inplace=True)

        # self.codon_table = self.codon_table.round(2)
        self.codon_table.index.name = 'Codon'
        self.codon_table = self.codon_table.reset_index(drop=False)
        self.codon_table.sort_values(by=['Abbr', 'Codon'], inplace=True)
        self.codon_table.to_csv(self.output + '_codon_occupancy.txt', header=True, index=False, sep='\t')

    def draw_occupancy_corr(self):
        out_pdf = self.output + "_occupancy_corrplot.pdf"
        out_png = self.output + "_occupancy_corrplot.png"

        occ_corr = self.codon_table[self.occ_title].astype('float').corr()
        occ_corr.to_csv(self.output + '_occupancy_corr.txt', sep='\t', index=True)

        matplotlib.use('AGG')
        fig = plt.figure(figsize=(7 + 0.2 * self.sample_num, 6 + 0.2 * self.sample_num), dpi=300)
        sns.heatmap(data=occ_corr, annot=True, fmt=".4f", cmap="vlag", linewidths=.01)
        plt.tight_layout()
        # plt.show()

        plt.savefig(fname=out_pdf)
        plt.savefig(fname=out_png)
        plt.close()

    def draw_occupancy_heat(self):
        # output the absolute occupancy heatmap
        out_pdf = self.output + "_occupancy_heatplot.pdf"
        out_png = self.output + "_occupancy_heatplot.png"

        matplotlib.use('AGG')
        fig = plt.figure(figsize=(10 + 0.5 * self.sample_num, 20), dpi=300)
        sns.heatmap(data=self.codon_table[self.occ_title].astype('float'),
                    yticklabels=self.codon_table['Codon'] + '[' + self.codon_table['Abbr'] + ']',
                    annot=True,
                    fmt=".3f",
                    cmap="vlag",
                    linewidths=.01)
        plt.tight_layout()
        # plt.show()

        plt.savefig(fname=out_pdf)
        plt.savefig(fname=out_png)
        plt.close()

    def draw_occupancy_relative_heat(self):

        # output the relative occupancy heatmap
        out_pdf = self.output + "_occupancy_relative_heatplot.pdf"
        out_png = self.output + "_occupancy_relative_heatplot.png"

        matplotlib.use('AGG')
        fig = plt.figure(figsize=(10 + 0.5 * self.sample_num, 20), dpi=300)
        sns.heatmap(data=self.codon_table[self.rel_title].astype('float'),
                    yticklabels=self.codon_table['Codon'] + '[' + self.codon_table['Abbr'] + ']',
                    annot=True,
                    fmt=".3f",
                    cmap="vlag",
                    linewidths=.01)
        plt.tight_layout()
        # plt.show()

        plt.savefig(fname=out_pdf)
        plt.savefig(fname=out_png)
        plt.close()

    def draw_occupancy_line(self):

        codon_occp = pd.melt(self.codon_table,
                             id_vars=['Codon', 'AA', 'Abbr'],
                             var_name='Groups',
                             value_name='Occupancy')
        codon_occp.loc[:, 'Codon'] = codon_occp['Codon'] + '[' + codon_occp['Abbr'] + ']'
        codon_occp = codon_occp.loc[codon_occp['Groups'].str.contains('relative'), ]

        out_pdf = self.output + "_occupancy_relative_lineplot.pdf"
        out_png = self.output + "_occupancy_relative_lineplot.png"

        # draw the odd ratio number of each codon
        matplotlib.use('AGG')

        fig, ax = plt.subplots(figsize=(11, 5.5), dpi=300)

        flag = 1
        for aa in codon_occp['AA'].unique().tolist():
            tmp = codon_occp.loc[codon_occp['AA'] == aa, :]
            if flag == 1:
                sns.lineplot(data=tmp,
                             x='Codon',
                             y='Occupancy',
                             linewidth=1.5,
                             hue='Groups',
                             style='Groups',
                             markers=True,
                             legend=True,
                             ax=ax)
                flag += 1
            else:
                sns.lineplot(data=tmp,
                             x='Codon',
                             y='Occupancy',
                             linewidth=1.5,
                             hue='Groups',
                             style='Groups',
                             markers=True,
                             legend=False,
                             ax=ax)

        # add the gene annotation here, need to import the txt file
        ax.legend(loc=8, ncol=4, bbox_to_anchor=(0.5, -0.5), fontsize=6)

        # ax.set_ylabel('occupancy')
        # ax.set_xlabel('codon')

        plt.xticks(rotation=90)

        plt.title('Relative codon occupancy')
        plt.tight_layout()
        # plt.show()

        fig.savefig(fname=out_pdf)
        fig.savefig(fname=out_png)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Project      :   riboparser
@Script       :   fa_length.py
@Environment  :   python 3.8.5
@Version      :   1.0
@Author       :   Rensc 
@Time         :   2023/05/29 11:26:28
@E-mail       :   rensc0718@163.com
@License      :   (C)Copyright 2023-2025, Ren Shuchao
'''


import argparse
from Bio import SeqIO
import gzip
import os


def args_parser():

    parser = argparse.ArgumentParser(description='This script is used to summarize the reads length.')

    # needed arguments
    input_group = parser.add_argument_group('Required arguments')
    input_group.add_argument('-i', dest='input', required=True, type=str, help='input file name')
    input_group.add_argument('-o', dest='output', required=True, type=str, help='output file name')

    args = parser.parse_args()
    args_dict = vars(args)
    for k, v in args_dict.items():
        print("{:<12}:  {:<}".format(k, str(v)), flush=True)

    return args


def import_reads(fa_file):
    '''
    @Message  :   import the fasta file
    @Input    :   fasta filename
    @Return   :   seqio record
    '''
    
    fa_suffix = os.path.splitext(fa_file)[1]
    if fa_suffix == '.gz':
        record = gzip.open(fa_file, "rt")
    else:
        record = open(fa_file, "r")

    return record


def summarize_reads(record):
    '''
    @Message  :   summarize the reads length
    @Input    :   SeqIO record
    @Return   :   a dict contain length and number
    
    {35: 103, 31: 904, 33: 38391, 32: 391495}
    '''
    
    length_dict = {}
    line = 0
    
    for reads in SeqIO.parse(record, "fasta"):
        line += 1
        if line % 1000000 == 0:
            print("rows : " + str(line), flush=True)

        length = len(reads)
        try:
            length_dict[length] += 1
        except KeyError:
            length_dict[length] = 1
    
    print("rows : " + str(line) + '\n', flush=True)

    return length_dict


def output_length_distr(length_dict, out_file):
    '''
    @Message  :   sort the length dict and output data
    @Input    :   dict contain length and number
    @Return   :   output the txt file
    '''
    length_dict = dict(sorted(length_dict.items(), key=lambda x: x[0]))

    with open(out_file, 'w') as out_flag:
        out_flag.writelines(''.join(['length', '\t', 'count', '\n']))

        for length, count in length_dict.items():
            out_flag.writelines(''.join([str(length), '\t', str(count), '\n']))


def main():
    print("Summarize the reads length of fasta file.", flush=True)
    args = args_parser()

    print("\nStep1: import the fasta file.", flush=True)
    record = import_reads(args.input)

    print("Step2: summarize the reads length.", flush=True)
    length_dict = summarize_reads(record)

    print("Step3: output the length distribution", flush=True)
    output_length_distr(length_dict, args.output)



if __name__ == '__main__':

    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Script  : get_win_seq.py
# @Time    : 2022/11/1
# @project : py-scripts
# @Version : python 3.7
# @Author  : Rensc
# @E-mail  : rensc0718@163.com


import argparse


def get_argparse():
    parser = argparse.ArgumentParser(description='rolling the windows on each sequence')
    parser.add_argument("-i", "--input", dest="fasta", required=False, help="input fasta")
    parser.add_argument("-w", "--win", dest="width", type=int, default=20, required=False, help="width of window")
    parser.add_argument("-o", "--output", dest="txt", required=False, help="output txt")
    args = parser.parse_args()
    return args


def get_win(seq, flag, win_width, outfile):
    length = len(seq)
    if length > win_width:
        b = [seq[i: i + win_width] for i in range(length + 1 - win_width)]
        for items in b:
            outfile.write('\t'.join([str(flag), seq, items]) + '\n')
    else:
        outfile.write('\t'.join([str(flag), seq, seq]) + '\n')


def read_file(inputfile, win_width, out_name):
    if out_name:
        pass
    else:
        out_name = inputfile + '.win'

    outfile = open(out_name, 'w')

    with open(inputfile, 'r') as file_in:
        flag = 0
        for line in file_in:
            if line.startswith('>'):
                continue
            flag += 1
            get_win(line.strip(), flag, win_width, outfile)

    outfile.close()


if __name__ == '__main__':
    args = get_argparse()
    read_file(args.fasta, args.width, args.txt)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project : riboParser
# @Script  : rpf_periodicity.py


from .ribo import ArgsParser
from .ribo import Periodicity

def main():
    ArgsParser.now_time()
    print('Draw the periodicity plot.\n', flush=True)
    print('Step1: Checking the input Arguments.\n', flush=True)
    args = ArgsParser.periodicity_args_parser()

    print('Step2: Import the RPFs file.\n', flush=True)
    rpfs = Periodicity.Periodicity(args)
    rpfs.import_rpf()

    print('Step3: Calculate the 3nt periodicity.\n', flush=True)
    rpfs.calc_3nt_period()

    print('Step4: Ouput the 3nt periodicity.\n', flush=True)
    rpfs.output_meta()

    print('Step5: Draw the 3nt periodicity plot.\n', flush=True)
    rpfs.draw_3nt_period_count()
    rpfs.draw_3nt_period_ratio()

    print('All done.\n', flush=True)
    ArgsParser.now_time()


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @time    : 2021/11/30
# @Project : riboParser
# @Script  : rpf_CDT.py


from .ribo import ArgsParser
from .ribo import CDT


def main():
    ArgsParser.now_time()
    print('\nCalculate the codon decoding time.', flush=True)
    print('\nStep1: Checking the input Arguments.', flush=True)
    args = ArgsParser.rpf_cdt_args_parser()
    cdt = CDT.CodonDecodingTime(args)

    print('\nStep2: Import the gene annotation.', flush=True)
    cdt.import_gene()

    print('\nStep3: Import the RPFs file.', flush=True)
    cdt.import_density()
    cdt.merge_rpf_rna()
    cdt.normalize_density()

    print('\nStep4: Calculate the codon decoding time.', flush=True)
    cdt.calc_rpf_norm()
    cdt.codon_decoding_time()

    print('\nStep5: Draw the codon decoding time plot.', flush=True)
    cdt.draw_cdt_corr()
    cdt.draw_cdt_heat()

    print('All done.\n', flush=True)
    ArgsParser.now_time()


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @time    : 2021/6/11 10:50
# @Project : riboParser
# @Script  : Digestion.py

import os
import re
from collections import OrderedDict
from itertools import islice
from Bio import SeqIO
from .Ribo import Mrna

import matplotlib
import matplotlib.pyplot as plt
import pandas as pd
import numpy as np
import pysam
import seaborn as sns

from Bio import motifs


class Ribo(object):

    def __init__(self, args):
        self.mrna_file = args.transcript
        self.mrna_seq = args.sequence
        self.mrna_dict = OrderedDict()
        self.longest = str(args.longest)

        self.bam = args.bam
        self.sample_format = os.path.splitext(args.bam)[1]
        self.pysam_input = None

        self.min_length = args.min
        self.max_length = args.max
        self.profile = None

        self.rpf_start = None
        self.rpf_end = None

        self.align_left = []
        self.align_right = []
        self.seqlogo_5end = None
        self.seqlogo_3end = None

        self.scaled = args.scale

        self.output = args.output

    def read_transcript(self):

        record = SeqIO.parse(self.mrna_seq, "fasta")
        mrna_sequence = OrderedDict()
        for line in record:
            # sys.stdout.writelines("import gene:  {gene}\n".format(gene=line.id))
            now_sequence = line.upper()
            if re.findall(r'[^ATGC]', str(now_sequence.seq)):
                print(line.id)
                print(now_sequence.seq)
                continue
            mrna_sequence[line.id] = now_sequence.seq

        with open(self.mrna_file, 'r') as trans_file_in:
            if self.longest:
                for line in islice(trans_file_in, 1, None):
                    record = line.strip().split('\t')

                    if record[9] == "True":
                        now_mrna = Mrna(record)
                        now_mrna.length = now_mrna.utr5_length + now_mrna.utr3_length + now_mrna.cds_length
                        try:
                            now_mrna.seq = mrna_sequence[
                                now_mrna.transcript_id]
                            now_mrna.rpf = [0] * len(now_mrna.seq)
                            self.mrna_dict[now_mrna.transcript_id] = now_mrna
                        except KeyError:
                            continue
                    else:
                        continue
            else:
                for line in islice(trans_file_in, 1, None):
                    record = line.strip().split('\t')
                    now_mrna = Mrna(record)

                    now_mrna.length = now_mrna.utr5_length + now_mrna.utr3_length + now_mrna.cds_length
                    try:
                        now_mrna.seq = mrna_sequence[now_mrna.transcript_id]
                        now_mrna.rpf = [0] * len(now_mrna.seq)
                        self.mrna_dict[now_mrna.transcript_id] = now_mrna
                    except KeyError:
                        continue

    def get_digest_sites(self):
        # import the bam file
        self.pysam_input = pysam.AlignmentFile(self.bam, "rb")
        # set the length range for the calculation
        if self.min_length < 20 and self.max_length > 100:
            rpf_start = {i: [0, 0, 0] for i in range(self.min_length, self.max_length + 1)}
            rpf_end = {i: [0, 0, 0] for i in range(self.min_length, self.max_length + 1)}
        elif self.min_length < 20:
            rpf_start = {i: [0, 0, 0] for i in range(self.min_length, 100 + 1)}
            rpf_end = {i: [0, 0, 0] for i in range(self.min_length, 100 + 1)}
        elif self.max_length > 100:
            rpf_start = {i: [0, 0, 0] for i in range(20, self.max_length + 1)}
            rpf_end = {i: [0, 0, 0] for i in range(20, self.max_length + 1)}
        else:
            rpf_start = {i: [0, 0, 0] for i in range(20, 100 + 1)}
            rpf_end = {i: [0, 0, 0] for i in range(20, 100 + 1)}

        for line in self.pysam_input.fetch(until_eof=True):
            if line.is_reverse:
                continue

            elif line.reference_name in self.mrna_dict:
                map_start, map_end = line.get_blocks()[0]
                read_length = line.infer_read_length()
                # check the location of rpf end
                mrna_length = self.mrna_dict[line.reference_name].length
                if map_start - 5 >= 0:
                    temp1 = self.mrna_dict[line.reference_name].seq[map_start - 5:map_start + 11]
                    if len(temp1) > 15:
                        self.align_left.append(temp1)
                if map_end + 5 < mrna_length:
                    # shift 1 nt for the alignment, (a,b] contain the map_end
                    temp2 = self.mrna_dict[line.reference_name].seq[map_end - 10:map_end + 6]
                    if len(temp2) > 15:
                        self.align_right.append(temp2)

                # filter the rpf with specific range
                if self.min_length <= read_length <= self.max_length:
                    frame_start = (map_start - self.mrna_dict[line.reference_name].utr5_length) % 3
                    frame_end = (map_end - 1 - self.mrna_dict[line.reference_name].utr5_length) % 3
                    rpf_start[read_length][frame_start] += 1
                    rpf_end[read_length][frame_end] += 1

        self.rpf_start = pd.DataFrame(rpf_start).T
        self.rpf_end = pd.DataFrame(rpf_end).T

    def output_digest_sites(self):
        psite_merge = pd.concat([self.rpf_start, self.rpf_end], join="outer", axis=1)
        psite_merge.fillna(0, inplace=True)
        psite_merge.columns = ["5end_frame1", "5end_frame2", "5end_frame3", "3end_frame1", "3end_frame2", "3end_frame3"]
        psite_merge.index.name = "length"
        psite_merge.to_csv(self.output + "_digestion_sites.txt", sep='\t', index=True)

    def digestion_plot(self):

        if self.scaled:
            out_pdf = self.output + "_scaled_digestion_sites_plot.pdf"
            out_png = self.output + "_scaled_digestion_sites_plot.png"

            psite_start_norm = self.rpf_start.sub(self.rpf_start.mean(axis=1), axis=0)
            psite_start_norm_row = psite_start_norm.div(psite_start_norm.std(axis=1), axis=0)
            psite_start_norm_row.fillna(0, inplace=True)

            psite_end_norm = self.rpf_end.sub(self.rpf_end.mean(axis=1), axis=0)
            psite_end_norm_row = psite_end_norm.div(psite_end_norm.std(axis=1), axis=0)
            psite_end_norm_row.fillna(0, inplace=True)

            now_cmap = "Blues"
        else:
            out_pdf = self.output + "_digestion_sites_plot.pdf"
            # out_png = self.output + "_digestion_sites_plot.png"

            psite_start_norm_row = self.rpf_start.copy()
            psite_end_norm_row = self.rpf_end.copy()

            now_cmap = "RdYlBu_r"

        codon_frame_5end = pd.DataFrame(self.rpf_start.loc[self.min_length:self.max_length].sum(axis=0)).T
        codon_frame_3end = pd.DataFrame(self.rpf_end.loc[self.min_length:self.max_length].sum(axis=0)).T
        rpfs_count_5end = pd.DataFrame(self.rpf_start.sum(axis=1)).loc[self.min_length:self.max_length].T
        rpfs_count_3end = pd.DataFrame(self.rpf_end.sum(axis=1)).loc[self.min_length:self.max_length].T

        matplotlib.use('Agg')
        # draw the periodicity of codon

        fig = plt.figure(figsize=(12, 10), dpi=300)
        ax1 = plt.subplot2grid((5, 6), (0, 0), rowspan=1, colspan=2)
        sns.barplot(data=codon_frame_5end, color="#388e70", alpha=.6, ax=ax1)
        ax1.set_xticklabels([])
        ax1.yaxis.set_ticks_position("right")
        ax1.set_ylabel("RPFs counts")
        ax1.set_title("5'-end")

        cbar_ax1 = fig.add_axes([0.43, 0.75, 0.01, 0.12])
        # ax11 = plt.subplot2grid((5, 6), (0, 2), rowspan=1, colspan=1)

        ax2 = plt.subplot2grid((5, 6), (1, 0), rowspan=4, colspan=2)
        sns.heatmap(
            data=psite_start_norm_row.loc[self.min_length:self.max_length],
            annot=None,
            linewidths=.5,
            ax=ax2,
            cmap=now_cmap,
            cbar_ax=cbar_ax1)
        label_y = ax2.get_yticklabels()
        plt.setp(label_y, rotation=0, horizontalalignment='right')
        ax2.set_ylabel("RPF length")
        ax2.set_xlabel("Codon frame")

        ax3 = plt.subplot2grid((5, 6), (1, 2), rowspan=4, colspan=1)
        sns.barplot(data=rpfs_count_5end,
                    color="#388e70",
                    alpha=.6,
                    ax=ax3,
                    orient='h')
        ax3.tick_params(bottom=True, top=False, left=False, right=False)
        ax3.set_yticklabels([])
        plt.xticks(rotation=60)
        ax3.set_xlabel("RPFs count")

        ax4 = plt.subplot2grid((5, 6), (0, 3), rowspan=1, colspan=2)
        sns.barplot(data=codon_frame_3end, color="#388e70", alpha=.6, ax=ax4)
        ax4.set_xticklabels([])
        ax4.yaxis.set_ticks_position("right")
        ax4.set_title("3'-end")

        cbar_ax2 = fig.add_axes([0.83, 0.75, 0.01, 0.12])
        ax5 = plt.subplot2grid((5, 6), (1, 3), rowspan=4, colspan=2)
        sns.heatmap(
            data=psite_end_norm_row.loc[self.min_length:self.max_length],
            annot=None,
            linewidths=.5,
            ax=ax5,
            cmap=now_cmap,
            cbar_ax=cbar_ax2)
        ax5.set_yticklabels([])
        ax5.set_xlabel("Codon frame")

        ax6 = plt.subplot2grid((5, 6), (1, 5), rowspan=4, colspan=1)
        sns.barplot(data=rpfs_count_3end,
                    color="#388e70",
                    alpha=.6,
                    ax=ax6,
                    orient='h')
        ax6.yaxis.set_ticks_position("right")
        plt.xticks(rotation=60)
        ax6.set_xlabel("RPFs count")

        # plt.tight_layout()
        plt.show()

        fig.savefig(fname=out_pdf)
        # fig.savefig(fname=out_png)

    def output_counts(self):
        self.seqlogo_5end = motifs.create(self.align_left)
        counts_5end = pd.DataFrame(self.seqlogo_5end.counts)
        pwm_5end = pd.DataFrame(self.seqlogo_5end.pwm)
        counts_5end.to_csv(self.output + "_5end_counts.txt", sep='\t', index=True)
        pwm_5end.to_csv(self.output + "_5end_pwm.txt", sep='\t', index=True)

        self.seqlogo_3end = motifs.create(self.align_right)
        counts_5end = pd.DataFrame(self.seqlogo_3end.counts)
        pwm_5end = pd.DataFrame(self.seqlogo_3end.pwm)
        counts_5end.to_csv(self.output + "_3end_counts.txt", sep='\t', index=True)
        pwm_5end.to_csv(self.output + "_3end_pwm.txt", sep='\t', index=True)

    def seq_logo_plot(self):
        import weblogo
        # get the 5'end digestion seqlogo
        seq_5end = weblogo.LogoData.from_counts(self.seqlogo_5end.alphabet,
                                                np.array(pd.DataFrame(self.seqlogo_5end.counts)))
        options = weblogo.LogoOptions()
        options.fineprint = '[-5, 10] from rpf 5-end digestion'
        format = weblogo.LogoFormat(seq_5end, options)
        pdf_5end = weblogo.pdf_formatter(seq_5end, format)
        with open(self.output + '_5end_seqlogo.pdf', 'wb') as f:
            f.write(pdf_5end)

        # get the 3'end digestion seqlogo
        seq_3end = weblogo.LogoData.from_counts(self.seqlogo_3end.alphabet,
                                                np.array(pd.DataFrame(self.seqlogo_3end.counts)))
        options = weblogo.LogoOptions()
        options.fineprint = '[-10, 5] from rpf 3-end digestion'
        format = weblogo.LogoFormat(seq_3end, options)
        pdf_3end = weblogo.pdf_formatter(seq_3end, format)
        with open(self.output + '_3end_seqlogo.pdf', 'wb') as f:
            f.write(pdf_3end)

    def seq_logo_plot2(self):
        import seqlogo
        # get the 5'end digestion seqlogo
        fname_5end = self.output + '_5end_seqlogo2.pdf'
        ppm_5end = seqlogo.Ppm(np.array([list(item) for item in self.seqlogo_5end.pwm.values()]))
        seqlogo.seqlogo(ppm_5end, ic_scale=False, format='pdf', size='medium', filename=fname_5end)
        seqlogo.seqlogo(ppm_5end, ic_scale=False, format='pdf', size='medium', filename=fname_5end)

        # get the 3'end digestion seqlogo
        fname_3end = self.output + '_3end_seqlogo2.pdf'
        ppm_3end = seqlogo.Ppm(np.array([list(item) for item in self.seqlogo_3end.pwm.values()]))
        seqlogo.seqlogo(ppm_3end, ic_scale=False, format='pdf', size='medium', filename=fname_3end)
        seqlogo.seqlogo(ppm_3end, ic_scale=False, format='pdf', size='medium', filename=fname_3end)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project : riboParser
# @Script  : rpf_merge.py


from collections import OrderedDict
import pandas as pd
import polars as pl

from .ribo import ArgsParser


def read_sample_list(sample_list):
    sp_list = pd.read_csv(sample_list, header=0, names=None, sep = '\t')

    sp_dict = OrderedDict()
    
    for _, rec in sp_list.iterrows():
        sp_dict[rec.iloc[0]] = rec.iloc[1]

    return sp_dict


def merge_rpf(sp_dict, output_prefix):
    total_rpf = None
    now_sample_num = 0
    for sp_name, sp_file in sp_dict.items():
        print('Import the RPFs file: {file_name}.'.format(file_name=sp_name), flush=True)
        if now_sample_num == 0:
            rpf = pl.read_csv(sp_file, separator='\t', has_header=True)
            rpf.columns = ["name", "now_nt", "from_tis", "from_tts", "region", "codon",
                           sp_name + "_f0", sp_name + "_f1", sp_name + "_f2"]
            total_rpf = rpf
        else:
            rpf = pl.read_csv(sp_file, separator='\t', has_header=True)
            rpf = rpf[:, [0, 1, 6, 7, 8]]
            rpf.columns = ["name", "now_nt", sp_name + "_f0", sp_name + "_f1", sp_name + "_f2"]
            total_rpf = total_rpf.join(rpf, on=['name', 'now_nt'])
        now_sample_num += 1

    output_name = output_prefix + "_merged.txt"
    total_rpf = total_rpf.with_columns(total_rpf['codon'].str.to_uppercase())
    total_rpf = total_rpf.filter(~total_rpf['codon'].str.contains('N'))

    total_rpf.write_csv(output_name, separator = '\t', include_header = True)


def main():
    ArgsParser.now_time()
    print('\nMerge RPFs files from different samples.', flush=True)
    print('\nStep1: Checking the input Arguments.', flush=True)
    args = ArgsParser.rpf_merge_args_parser()

    print('\nStep2: Import the sample list.', flush=True)
    sp_list = read_sample_list(args.list)

    print('\nStep3: Merge the RPFs file.', flush=True)
    merge_rpf(sp_list, args.output)

    print('\nAll done.', flush=True)
    ArgsParser.now_time()


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
#@Project : riboParser
#@Script  : merge_saturation.py


import os
import pandas as pd
from collections import OrderedDict
import argparse


def merge_saturation_args_parser():

    parser = argparse.ArgumentParser(description='This script is used to merge the saturation files')

    # needed arguments
    input_group = parser.add_argument_group('Required arguments')
    input_group.add_argument(
        "-l", "--list", nargs='+', required=True, help="List for saturation files (e.g., '*_gene_saturation.txt')."
    )
    input_group.add_argument(
        '-o', dest='output', required=True, type=str, help='prefix of output file name.'
    )

    args = parser.parse_args()
    args_dict = vars(args)
    for k, v in args_dict.items():
        print('{:<12}:  {:<}'.format(k, str(v)), flush=True)

    return args


def process_saturation_files(gene_count_list):
    '''
    @Message  : retrieve the length table.
    @Input    : list --> pattern for gene saturation files (e.g., '*_gene_saturation.txt')
    @Return   : 
                output --> output nested dict contain the gene saturation
    @Flow     : step1 --> retrieve the gene saturation
    '''

    # make the dataframe with title
    gene_count_df = pd.DataFrame(columns=['Sample', 'Tercile', 'Covered', 'Uncovered'])

    # for each gne file
    for gene_file in gene_count_list:
        file_prefix = os.path.basename(gene_file).split('.')[0].replace('_gene_saturation', '')

        # import the gene saturation file
        gene_df = pd.read_csv(gene_file, sep='\t', index_col=False)
        gene_df.columns = ['Tercile', 'Covered']
        gene_df['Uncovered'] = gene_df['Covered'].max() - gene_df['Covered']
        gene_df['Sample'] = file_prefix
        gene_df = gene_df[['Sample', 'Tercile', 'Covered', 'Uncovered']]

        # merge the gene saturation file
        gene_count_df = pd.concat([gene_count_df, gene_df], axis=0)

    return gene_count_df


def output_table(gene_count_df, output_file):
    '''
    @Message  : function for output.
    @Input    : result_dict --> nested dict contain the gene count
    @Return   : output --> description
    @Flow     : step1 --> convert the number to int
    '''

    # keep the 2 decimal places
    gene_count_df[['Tercile', 'Covered', 'Uncovered']] = gene_count_df[['Tercile', 'Covered', 'Uncovered']].astype('int')

    gene_count_df.to_csv(output_file + '_gene_saturation.txt', sep='\t', index=False)


def main():

    print('Step1: Checking the input Arguments.', flush=True)
    args = merge_saturation_args_parser()

    print('Step2: import the gene saturation file.', flush=True)
    gene_count_df = process_saturation_files(args.list)

    print('Step3: output the gene saturation table.', flush=True)
    output_table(gene_count_df, args.output)

    print('All done.', flush=True)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project : riboParser
# @Script  : make_ensb_ref.py


import sys

import numpy as np
from Bio import SeqIO

from .ribo import ArgsParser
from .ribo.Ensembl_Ref import *


def readline(record):
    chrom, source, feature, start, end, score, strand, phase, attr = record.split('\t')

    attr_dict = OrderedDict()
    for num, mess in enumerate(attr.strip(';').split(';')):
        idx, values = mess.strip().split(' "')
        attr_dict[idx] = values.strip('"')

    section = {"chrom": chrom, "source": source, "feature": feature,
               "start": int(start), "end": int(end), "strand": strand, "attr": attr,
               "attr_dict": attr_dict, "mess": record}

    return section


def detect_cds_periodicity(gtf_filename):
    # CDS frame in gtf file could be fit the 3nt periodicity, commonly, the frame of CDS are Closed-interval
    # but some times the frame is Right-open-interval
    transcripts_dict = OrderedDict()
    now_mrna = 0
    with open(gtf_filename, 'r') as gtf_in:
        for line in gtf_in:
            record = line.strip()
            if not record or line.startswith('#'):
                continue

            section = readline(record)
            if section["attr_dict"]["gene_biotype"] == "protein_coding":
                if section["feature"] == "transcript":
                    transcripts_dict[section["attr_dict"]["transcript_id"]] = np.array([0, 0, 0])
                    now_mrna += 1
                    if now_mrna >= 2000:
                        sys.stdout.writelines("The first 1000 genes was used to detect the format of CDS position.\n")
                        break
                elif section["feature"] == "CDS":
                    cds_frame0 = section["end"] - section["start"]
                    cds_frame1 = section["end"] - section["start"] + 1
                    cds_frame2 = section["end"] - section["start"] - 1
                    transcripts_dict[section["attr_dict"]["transcript_id"]] += [cds_frame0, cds_frame1, cds_frame2]

    cds_shift = 1
    cds_type = ["Right-open-interval", "Closed-interval", "Open-interval"]
    for frame in [0, 1, -1]:
        mrna_length_array = np.array(list(map(lambda length: length[frame], transcripts_dict.values())))

        cds_frame = mrna_length_array % 3
        out_frame_num = np.count_nonzero(cds_frame)
        if out_frame_num == 0:
            # print(transcripts_dict)
            sys.stdout.writelines("The frame type of CDS is {now_type}.\n".format(now_type=cds_type[frame]))
            cds_shift = frame
            break
        elif frame == -1:
            sys.stdout.writelines("Some CDS in GTF file does not fit to 3 nt periodicity.\n")
            cds_shift = 1
        else:
            continue

    return cds_shift


def read_gtf(gtf_filename):
    title_list = []
    genes_dict = OrderedDict()
    transcripts_dict = OrderedDict()
    now_row = 0

    cds_shift = detect_cds_periodicity(gtf_filename)

    with open(gtf_filename, 'r') as gtf_in:
        for line in gtf_in:
            now_row += 1
            if now_row % 10000 == 0:
                sys.stdout.writelines("Rows:  {number}\n".format(number=now_row))

            record = line.strip()

            # skip the '#' lines and blank lines
            if not record:
                continue
            if line.startswith('#'):
                title_list.append(record)
                continue

            section = readline(record)

            # merge the gene lines
            if section["feature"] == "gene":
                genes_dict[section["attr_dict"]["gene_id"]] = Gene(section)

            # merge the mRNA
            elif section["feature"] == "transcript":
                now_rna = Transcripts(section)
                transcripts_dict[now_rna.transcript_id] = now_rna

            # merge the exon, cds, start_codon, stop_codon
            elif section["feature"] in ["exon", "CDS", "start_codon", "stop_codon"]:
                trans_id = section["attr_dict"]["transcript_id"]
                transcripts_dict[trans_id].add_feature(section, cds_shift)

            # skip the other genes
            else:
                # sys.stdout.write("Skip: {row} ".format(row=record))
                continue

    sys.stdout.writelines("Rows: {number}\n".format(number=now_row))

    return title_list, genes_dict, transcripts_dict


def gene_tree(utr_len, chroms_dict, transcripts_dict, genes_dict):
    for trans_id, trans_info in transcripts_dict.items():

        if trans_info.gene_id in genes_dict:
            if trans_info.transcript_biotype == "protein_coding":
                try:
                    trans_info.add_utr(utr_len, chroms_dict)
                except IndexError:
                    sys.stdout.write("IndexError: {gene}\n".format(gene=trans_info.gene_id))
                    continue

            genes_dict[trans_info.gene_id].add_transcript(trans_info)

        else:
            raise KeyError("Error: {trans} not found in {gene}!".format(trans=trans_id, gene=trans_info.gene_id))

    return genes_dict


def format_results(gene_mess):
    mrna_gtf = []
    gene_gtf = []
    mrna_txt = []
    mrna_region = []

    gene_gtf.append('\t'.join([gene_mess.chrom, gene_mess.source, gene_mess.feature, str(gene_mess.start),
                               str(gene_mess.end), '.', gene_mess.strand, '.', gene_mess.attr]))

    if gene_mess.gene_type == "protein_coding":
        mrna_gtf.append('\t'.join([gene_mess.chrom, gene_mess.source, gene_mess.feature, str(gene_mess.start),
                                   str(gene_mess.end), '.', gene_mess.strand, '.', gene_mess.attr]))

    for trans_ids, trans_info in gene_mess.transcript.items():
        if trans_info.transcript_biotype == "protein_coding":
            # Determine whether the CDS length is an integer multiple of 3.
            if trans_info.cds_length % 3 != 0:
                sys.stdout.write("Warning! {gene} CDS length doesn't fit the 3nt periodicity. \n".format(gene=trans_ids))
                continue

            if gene_mess.rep_transcript == trans_info.transcript_id:
                rep_transcript = True
            else:
                rep_transcript = False

            mrna_txt.append('\t'.join([trans_info.chrom, gene_mess.gene_id, trans_info.gene_name, trans_info.transcript_id,
                                       str(trans_info.start), str(trans_info.end), str(trans_info.utr5), str(trans_info.cds_length),
                                       str(trans_info.utr3), trans_info.strand, str(rep_transcript), str(trans_info.modified)]))
            mrna_region.append([trans_ids, trans_info.chrom, trans_info.exons, trans_info.strand])

            if not trans_info.modified:
                mrna_gtf.extend(trans_info.mess)
                gene_gtf.extend(trans_info.mess)

            else:
                mrna_gtf.append('\t'.join([trans_info.chrom, trans_info.source, trans_info.feature, str(trans_info.start),
                                           str(trans_info.end), '.', trans_info.strand, '.', trans_info.attr]))
                gene_gtf.append('\t'.join([trans_info.chrom, trans_info.source, trans_info.feature, str(trans_info.start),
                                           str(trans_info.end), '.', trans_info.strand, '.', trans_info.attr]))

                for exon in trans_info.exon_feature:
                    mrna_gtf.append('\t'.join([str(i) for i in exon]))
                    gene_gtf.append('\t'.join([str(i) for i in exon]))

                for cds in trans_info.cds_feature:
                    mrna_gtf.append('\t'.join([str(i) for i in cds]))
                    gene_gtf.append('\t'.join([str(i) for i in cds]))
        else:
            gene_gtf.extend(trans_info.mess)

    return mrna_gtf, gene_gtf, mrna_txt, mrna_region


def filter_genes(genes_dict, coding):
    filtered_gtf = []
    filtered_txt = []
    filtered_region = []

    if coding:
        for gene_name, gene_mess in genes_dict.items():
            mrna_gtf, gene_gtf, mrna_txt, mrna_region = format_results(gene_mess)
            filtered_gtf.extend(mrna_gtf)
            filtered_txt.extend(mrna_txt)
            filtered_region.extend(mrna_region)

    elif not coding:
        for gene_name, gene_mess in genes_dict.items():
            if gene_mess.gene_type == 'protein_coding':
                mrna_gtf, gene_gtf, mrna_txt, mrna_region = format_results(gene_mess)
                filtered_gtf.extend(gene_gtf)
                filtered_txt.extend(mrna_txt)
                filtered_region.extend(mrna_region)
            else:
                mrna_gtf, gene_gtf, mrna_txt, mrna_region = format_results(gene_mess)
                filtered_gtf.extend(gene_gtf)

    return filtered_gtf, filtered_txt, filtered_region


def read_genome(genome):
    chroms_dict = OrderedDict()

    record = SeqIO.parse(genome, "fasta")
    for line in record:
        sys.stdout.writelines("import chromosome: {chrom}\n".format(chrom=line.id))
        chroms_dict[line.id] = Chrom(line)

    return chroms_dict


def get_seq(chroms_dict, mrna_region):
    mrna_seq = OrderedDict()

    for transcript in mrna_region:
        transcript_seq = ''

        if transcript[-1] == "-":
            for exon in reversed(transcript[2]):
                exon_start, exon_end = exon[0], exon[1]
                transcript_seq += chroms_dict[transcript[1]].seq[exon_start - 1: exon_end]
            transcript_seq = transcript_seq.reverse_complement()

        else:
            for exon in transcript[2]:
                exon_start, exon_end = exon[0], exon[1]
                transcript_seq += chroms_dict[transcript[1]].seq[exon_start - 1: exon_end]

        mrna_seq[transcript[0]] = transcript_seq

    return mrna_seq


def output_results(output_prefix, title_list, filtered_gtf, mrna_txt, mrna_seq):
    gtf_out_file = output_prefix + '.norm.gtf'
    with open(gtf_out_file, 'w') as gtf_out:
        for line in title_list:
            gtf_out.writelines(''.join(line) + '\n')

        for line in filtered_gtf:
            gtf_out.writelines(''.join(line) + '\n')

    txt_out_file = output_prefix + '.norm.txt'
    with open(txt_out_file, 'w') as txt_out:
        txt_out.writelines('\t'.join(["chromosome", "gene_id", "gene_name", "transcript_id", "start", "end", "utr5_length",
                                      "cds_length", "utr3_length", "strand", "rep_transcript", "modified"]) + '\n')
        for line in mrna_txt:
            txt_out.writelines(''.join(line) + '\n')

    seq_out_file = output_prefix + '.norm.fa'
    with open(seq_out_file, 'w') as seq_out:
        for mrna, sequence in mrna_seq.items():
            seq_out.writelines('\n'.join([">" + mrna, str(sequence)]) + '\n')


def main():
    ArgsParser.now_time()
    sys.stdout.writelines('\nMake the gene annotation files.\n')
    sys.stdout.writelines('Step1: Checking the input Arguments.\n')
    args = ArgsParser.gtf_args_parser()

    sys.stdout.writelines('\nStep2: Import the gtf file.\n')
    title_list, genes_dict, transcripts_dict = read_gtf(args.transcript)

    sys.stdout.writelines('\nStep3: Import the genome file.\n')
    chroms_dict = read_genome(args.sequence)

    sys.stdout.writelines('\nStep4: Make the gene tree.\n')
    genes_dict = gene_tree(args.utr, chroms_dict, transcripts_dict, genes_dict)

    sys.stdout.writelines('\nStep5: Screening genes.\n')
    filtered_gtf, mrna_txt, mrna_region = filter_genes(genes_dict, args.coding)

    sys.stdout.writelines('\nStep6: Retrieve the mRNA sequence from genome.\n')
    mrna_seq = get_seq(chroms_dict, mrna_region)

    sys.stdout.writelines('\nStep7: Output the results.\n')
    output_results(args.output, title_list, filtered_gtf, mrna_txt, mrna_seq)

    sys.stdout.writelines('\nALL DONE!\n\n')
    ArgsParser.now_time()


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project : riboParser
# @Script  : rpf_Check.py


from .ribo import ArgsParser
from .ribo.Quality import *


def main():
    ArgsParser.now_time()
    print('Check the RPFs mapping condition.\n', flush=True)
    print('Step1: Checking the input Arguments.\n', flush=True)
    args = ArgsParser.rpf_bam_check_parser()
    rpf_quality = Quality(args)

    print('Step2: Import the transcripts annotation.\n', flush=True)
    rpf_quality.read_transcript()

    print('Step3: Import the bam file.\n', flush=True)
    rpf_quality.sort_index_bam()
    rpf_quality.fliter_mrna_reads()

    print('Step4: Sort and index the bam file.\n', flush=True)
    rpf_quality.merge_sort_index_bam()

    print('Step5: Detect the type of sequence profile.\n', flush=True)
    if not rpf_quality.profile:
        rpf_quality.detect_seq_type()
    else:
        print("{bam} is specified as {ribo}-seq.\n".format(bam=rpf_quality.sample_file, ribo=rpf_quality.profile),
              flush=True)

    print('Step6: Summary the length distribution of reads aligned to mRNA.\n', flush=True)
    rpf_quality.write_length_distr()

    if args.saturation:
        print('Step7: Check the RPFs saturation.\n', flush=True)
        rpf_quality.rpf_saturation()
        rpf_quality.draw_gene_saturation()
        rpf_quality.draw_rpf_saturation()
    else:
        print('Step7: Don not check the RPFs saturation.\n', flush=True)

    print('All done.', flush=True)
    ArgsParser.now_time()


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Project      : riboParser
@Script       : fa_len_flt.py
@Environment  : python 3.8.5
@Version      : 1.0
@Author       : Rensc 
@Time         : 2023/11/29 15:44:09
@E-mail       : rensc0718@163.com
@License      : (C)Copyright 2023-2025, Ren Shuchao
'''

import argparse
import os
import gzip
from Bio import SeqIO


# get arguments and define the scripts usage
def args_parser():
    parser = argparse.ArgumentParser(description='This script is used to filter the reads length.')

    # needed arguments
    input_group = parser.add_argument_group('Required arguments')
    input_group.add_argument('-i', dest='input', required=True, type=str, help='input fasta file name')
    input_group.add_argument('-o', dest='output', required=True, type=str, help='output fasta file name')

    # optional arguments
    optional_group = parser.add_argument_group('Optional arguments')
    optional_group.add_argument('-m', dest='min', type=int, default=8, help='min length of reads. (default %(default)s)')
    optional_group.add_argument('-M', dest='max', type=int, default=1000, help='max length of reads. (default %(default)s)')

    args = parser.parse_args()
    args_dict = vars(args)
    for k, v in args_dict.items():
        print("{:<12}:  {:<}".format(k, str(v)), flush=True)

    return args


def import_reads(fa_file):
    '''
    @Message  :   import the fasta file
    @Input    :   fasta filename
    @Return   :   seqio record
    '''
    
    fa_suffix = os.path.splitext(fa_file)[1]
    if fa_suffix == '.gz':
        record = gzip.open(fa_file, "rt")
    else:
        record = open(fa_file, "r")

    return record


def filter_reads(record, output, min_length, max_length):
    '''
    @Message  :   filter the reads length
    @Input    :   SeqIO record
    @Return   :   output the fasta file contain reads length between min and max
    
    '''

    with open(output, "w") as out_flag:
        line = 0
        for reads in SeqIO.parse(record, "fasta"):
            line += 1
            if line % 1000000 == 0:
                print("rows : " + str(line), flush=True)

            length = len(reads)
            if length >= min_length and length <= max_length:
                SeqIO.write(reads, out_flag, "fasta")


def main():
    print("Filter the reads length of fasta file.", flush=True)
    args = args_parser()

    print("\nStep1: import the fasta file.", flush=True)
    record = import_reads(args.input)

    print("Step2: output the fasta.", flush=True)
    filter_reads(record, args.output, args.min, args.max)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
#@Project : riboParser
#@Script  : merge_cdt.py


import pandas as pd
import argparse


def merge_cdt_args_parser():

    parser = argparse.ArgumentParser(description='This script is used to merge the codon decoding time files')

    # needed arguments
    input_group = parser.add_argument_group('Required arguments')
    input_group.add_argument(
        "-l", "--list", nargs='+', required=True, help="List for cdt files (e.g., '*_cdt.txt')."
    )
    input_group.add_argument(
        '-o', dest='output', required=True, type=str, help='prefix of output file name (default: prefix + _cdt.txt).'
    )

    args = parser.parse_args()
    args_dict = vars(args)
    for k, v in args_dict.items():
        print('{:<12}:  {:<}'.format(k, str(v)), flush=True)

    return args


def import_cdt_files(cdt_list):
    '''
    @Message  : retrieve the cdt table.
    @Input    : list --> pattern for cdt files (e.g., '*_codon_cdt.txt')
    @Return   : 
                output --> output dataframe contain the cdt
    @Flow     : step1 --> retrieve the cdt
    '''

    # make the dataframe
    cdt_df = pd.DataFrame()

    # for each cdt file
    for cdt_file in cdt_list:
        # import the cdt file
        cdt = pd.read_csv(cdt_file, sep='\t', index_col=False)
        cdt.sort_values(by=['Abbr', 'Codon'], ascending=True, inplace=True, ignore_index=True)

        # merge the gene cdt file
        if cdt_df.empty:
            cdt_df = cdt
        else:
            # delete the columns ['AA', 'Abbr.']
            cdt = cdt.drop(['AA', 'Abbr'], axis=1)
            cdt_df = pd.merge(left=cdt_df, right=cdt, how='outer', on='Codon')

    # reorder the cdt columns with 
    # ['codon', 'AA', 'Abbr.', *_density, *_absolute_cdt, *_relative_cdt]
    cdt_codon = cdt_df[['Codon', 'AA', 'Abbr']]
    absolute_cdt = cdt_df.filter(regex='_absolute_cdt')
    relative_cdt = cdt_df.filter(regex='_relative_cdt').filter(regex='^(?!.*norm).*$')
    norm_cdt = cdt_df.filter(regex='_norm_cdt')
    relative_norm_cdt = cdt_df.filter(regex='_norm_relative_cdt')

    cdt_merge = pd.concat([cdt_codon, absolute_cdt, norm_cdt, relative_cdt, relative_norm_cdt], axis=1)

    return cdt_merge


def output_table(cdt_merge, output_file):
    '''
    @Message  : function for output.
    @Input    : result_dict --> nested dict contain the cdt
    @Return   : output --> description
    @Flow     : step1 --> convert the dict to dataframe
    '''

    cdt_merge.to_csv(output_file + '_codon_decoding_time.txt', sep='\t', index=False)


def main():

    print('Step1: Checking the input Arguments.', flush=True)
    args = merge_cdt_args_parser()

    print('Step2: import the cdt file.', flush=True)
    cdt_df = import_cdt_files(args.list)

    print('Step3: output the cdt table.', flush=True)
    output_table(cdt_df, args.output)

    print('All done.', flush=True)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project : riboParser
# @Script  : RNA.py


import os
import sys
from collections import OrderedDict
from itertools import islice
from Bio import SeqIO
import numpy as np
import pandas as pd
import polars as pl
import pysam


class Mrna(object):

    def __init__(self, record):
        self.chromosome = record[0]
        self.gene_id = record[1]
        # self.gene_name = record[2]
        self.transcript_id = record[2]
        self.start = record[3]
        self.end = record[4]
        self.utr5_length = int(record[5])
        self.cds_length = int(record[6])
        self.utr3_length = int(record[7])
        self.strand = record[8]
        self.rep_transcript = record[9]
        self.modified = record[10]
        self.bam = []
        self.reads = []
        self.seq = None


class RNA(object):

    def __init__(self, args):
        self.mrna_file = args.transcript
        self.mrna_seq = args.sequence
        self.mrna_dict = OrderedDict()
        self.longest = args.longest

        self.bam = args.bam
        self.sample_format = os.path.splitext(args.bam)[1]
        self.pysam_input = None

        self.psite = args.psite
        self.min_length = args.min
        self.max_length = args.max

        self.offset = {}
        self.profile = None
        self.rolling = args.rolling
        self.pair_end = args.pair_end

        self.thread = args.thread

        self.output = args.output

    def read_transcript(self):
        '''
        @Message  : read the transcript file
        @Input    : self.mrna_file --> mrna message table generated by riboParser
                    self.mrna_seq --> mrna sequence file
        @Return   : self.mrna_dict --> mrna dict contain the mrna message
        @Flow     : step1 --> read the mrna sequence file
                    step2 --> read the mrna message file
                    step3 --> filter the longest transcript
                    
        '''
        
        # read the mrna sequence file
        record = SeqIO.parse(self.mrna_seq, "fasta")
        mrna_sequence = OrderedDict()
        for line in record:
            # sys.stdout.writelines("import gene:  {gene}\n".format(gene=line.id))
            mrna_sequence[line.id] = line.seq

        # read the mrna message file
        with open(self.mrna_file, 'r') as trans_file_in:
            if self.longest:
                for line in islice(trans_file_in, 1, None):
                    record = line.strip().split('\t')
                    # if int(record[7]) % 3 != 0:
                    #     sys.stdout.write("Delete {gene}. CDS length doesn't fit 3nt periodicity. \n".format(gene=record[3]))

                    if record[9] == "True":
                        now_mrna = Mrna(record)
                        # sys.stdout.writelines(now_mrna.transcript_id + '\n')
                        now_mrna.length = now_mrna.utr5_length + now_mrna.utr3_length + now_mrna.cds_length
                        try:
                            now_mrna.seq = mrna_sequence[now_mrna.transcript_id]
                            now_mrna.reads = np.zeros(len(now_mrna.seq))
                            self.mrna_dict[now_mrna.transcript_id] = now_mrna
                        except KeyError:
                            continue
                    else:
                        continue
            else:
                for line in islice(trans_file_in, 1, None):
                    record = line.strip().split('\t')
                    # if int(record[7]) % 3 != 0:
                    #     sys.stdout.write("Delete {gene}. CDS length doesn't fit 3nt periodicity. \n".format(gene=record[3]))
                    
                    now_mrna = Mrna(record)
                    # sys.stdout.writelines(now_mrna.transcript_id + '\n')
                    now_mrna.length = now_mrna.utr5_length + now_mrna.utr3_length + now_mrna.cds_length
                    try:
                        now_mrna.seq = mrna_sequence[now_mrna.transcript_id]
                        now_mrna.reads = np.zeros(len(now_mrna.seq))
                        self.mrna_dict[now_mrna.transcript_id] = now_mrna
                    except KeyError:
                        continue

    def read_offset(self):
        '''
        @Message  : read the offset file
        @Input    : self.psite --> input offset file name
        @Return   : self.offset --> offset dictionary
        @Flow     : step1 --> read the offset file
                    step2 --> drop the null value
                    step3 --> change the offset type to int
        '''
        
        offset = pd.read_csv(self.psite, sep='\t', header=0, names=None)
        offset.dropna(axis=0, inplace=True)
        offset["p_site"] = offset["p_site"].astype(int)
        offset["p_site"] = offset["p_site"] - 1
        self.offset = offset[["length", "p_site"]].groupby(offset["length"])["p_site"].apply(list).T.to_dict()
        print(self.offset)

    def read_bam(self):
        '''
        @Message  : read the bam file
        @Input    : self.bam --> input bam file name
        @Return   : self.pysam_input --> pysam.AlignmentFile
        @Flow     : step1 --> check the input file format
                    step2 --> read the bam file
                    step3 --> check the bam index
                    step4 --> build the bam index
        '''
        
        # check the input file format
        if self.sample_format.lower() == ".bam":
            sys.stdout.writelines("import file: {bam}.\n".format(bam=self.bam))
            self.read_format = 'rb'
        elif self.sample_format.lower() == ".sam":
            sys.stdout.writelines("import file: {bam}.\n".format(bam=self.bam))
            self.read_format = 'r'
        else:
            sys.stdout.writelines("Unknown file format, please input the correct bam or sam file.")
            sys.exit()

        self.pysam_input = pysam.AlignmentFile(self.bam, self.read_format)
        
        # check the bam index
        if self.pysam_input.has_index():
            pass
        else:
            import subprocess
            print("build the bam index for : {bam}".format(bam=self.bam), flush=True)

            subprocess.call("samtools index {bam}".format(bam=self.bam), shell=True)

            self.pysam_input = pysam.AlignmentFile(self.bam, self.read_format)

    def mono_density(self):
        '''
        @Message  : calculate the reads monosome density
        @Input    : self.mrna_dict --> mrna dict contain the mrna message
                    self.pysam_input --> pysam.AlignmentFile
        @Return   : self.mrna_dict --> mrna dict contain the mrna message and reads density
        @Flow     : step1 --> delete the unmapped reads
                    step2 --> delete the reverse reads
                    step3 --> filter the reads length
                    step4 --> count the reads density
        '''
        
        rows = 0

        # fetch the reads from bam file
        for line in self.pysam_input.fetch(until_eof=True):
            rows += 1
            if rows % 1000000 == 0:
                print('rows : ' + str(rows), flush=True)

            # delete the unmapped reads
            if line.is_unmapped:
                continue

            # delete the reverse reads
            if line.is_reverse and not self.pair_end:
                continue

            # count the reads density
            if line.reference_name in self.mrna_dict:
                map_start, map_end = line.get_blocks()[0]
                read_length = line.infer_read_length()

                if (self.min_length <= read_length <= self.max_length) and (read_length in self.offset):
                    
                    if self.offset[read_length][0] == -1:
                        p_site = line.get_reference_positions()
                        self.mrna_dict[line.reference_name].reads[p_site] += 1

                    else:
                        p_site = map_start + self.offset[read_length][0]
                        self.mrna_dict[line.reference_name].reads[p_site] += 1
                else:
                    pass
        
        print('rows : ' + str(rows), flush=True)

        self.pysam_input.close()

    def roll_density(self):
        '''
        @Message  : calculate the reads rolling polysome density with multi-thread
        @Input    : self.mrna_dict --> mrna dict contain the mrna message
                    self.pysam_input --> pysam.AlignmentFile
        @Return   : self.mrna_dict --> mrna dict contain the mrna message and reads density
        @Flow     : step1 --> delete the unmapped reads
                    step2 --> delete the reverse reads
                    step3 --> filter the reads length
                    step4 --> count the reads density with rolling polysome
        '''
        
        rows = 0
        # fetch the reads from bam file
        for line in self.pysam_input.fetch(until_eof=True):
            rows += 1
            if rows % 1000000 == 0:
                print('rows : ' + str(rows), flush=True)

            # delete the unmapped reads
            if line.is_unmapped:
                continue

            # delete the reverse reads
            if line.is_reverse and not self.pair_end:
                continue

            if line.reference_name in self.mrna_dict:
                map_start, map_end = line.get_blocks()[0]
                read_length = line.infer_read_length()
                
                if (self.min_length <= read_length <= self.max_length) and (read_length in self.offset):

                    if self.offset[read_length][0] == -1:
                        p_site = line.get_reference_positions()
                        self.mrna_dict[line.reference_name].reads[p_site] += 1

                    else:
                        p_site_list = np.arange(map_start, map_end, 32)
                        self.mrna_dict[line.reference_name].reads[p_site_list] += 1

                else:
                    pass
        
        print('rows : ' + str(rows), flush=True)

        self.pysam_input.close()

    def calculate_density(self):
        '''
        @Message  : function for calculate the reads density
        @Input    : self.rolling --> bool, True for rolling polysome, False for monosome
        @Return   : self.mrna_dict --> mrna dict contain the mrna message and reads density
        @Flow     : 
        '''

        # run the mono_density or roll_density function
        if self.rolling:
            self.roll_density()
        else:
            self.mono_density()

    @staticmethod
    def format_reads(args):
        '''
        @Message  : function for format the reads density
        @Input    : self.mrna_dict --> mrna dict contain the mrna message and reads density
        @Return   : self.total_reads_df --> dataframe contain the reads density
        @Flow     : step1 --> trim the cds frame to fit the 3nt periodicity
                    step2 --> trim the reads density
                    step3 --> format the reads density
                    step4 --> merge the reads density
        '''
        
        gene_list, mrna_dict, out_prefix = args
        reads_list = []

        for name in gene_list:
            # retrieve the gene message from the reads density
            isoform = mrna_dict[name]
            reads_dict = OrderedDict()

            # trim the cds frame to fit the 3nt periodicity
            if isoform.cds_length % 3 != 0:
                shift_cds = isoform.cds_length % 3
                isoform.cds_length -= shift_cds
                isoform.utr3_length += shift_cds

            shift_5 = isoform.utr5_length % 3
            shift_3 = isoform.utr3_length % 3

            utr5_length = isoform.utr5_length - shift_5
            utr3_length = isoform.utr3_length - shift_3
            trim_length = isoform.length - shift_5 - shift_3

            # trim the reads density
            reads_dict['name'] = [name] * int(trim_length / 3)
            reads_dict['now_nt'] = list(range(1 + shift_5, isoform.length - shift_3, 3))
            reads_dict['from_tis'] = list(range(-utr5_length // 3, (trim_length - utr5_length) // 3))
            reads_dict['from_tts'] = list(range((utr3_length - trim_length) // 3 + 1, utr3_length // 3 + 1))
            reads_dict['region'] = ['5utr'] * (utr5_length // 3) + ['cds'] * (isoform.cds_length // 3) + ['3utr'] * (utr3_length // 3)

            # trim the split gene nucleotide sequence to codon
            if shift_3 != 0:
                trim_seq = isoform.seq[shift_5: -shift_3]
                reads_dict['codon'] = [str(trim_seq[i:i + 3]) for i in range(0, trim_length, 3)]

                trim_reads = isoform.reads[shift_5: -shift_3]
                reads_dict[out_prefix + '_f0'] = trim_reads[0::3]
                reads_dict[out_prefix + '_f1'] = trim_reads[1::3]
                reads_dict[out_prefix + '_f2'] = trim_reads[2::3]
            elif shift_3 == 0:
                trim_seq = isoform.seq[shift_5::]
                reads_dict['codon'] = [str(trim_seq[i:i + 3]) for i in range(0, trim_length, 3)]

                trim_reads = isoform.reads[shift_5::]
                reads_dict[out_prefix + '_f0'] = trim_reads[0::3]
                reads_dict[out_prefix + '_f1'] = trim_reads[1::3]
                reads_dict[out_prefix + '_f2'] = trim_reads[2::3]
            try:
                reads = pd.DataFrame(reads_dict)
                reads_list.append(reads)
            except ValueError:
                print('Skip the genes: ' + str(name), flush=True)

        total_reads_df = pd.concat(reads_list, ignore_index=True)
        return total_reads_df

    def run_format_reads_with_multi_thread(self):
        """
        1. split the gene list into several sub-list
        2. run the format_rreads_density function with multi-thread
        3. merge the long data from each sub-list

        """
        
        # split the gene list into several sub-list
        gene_list = list(self.mrna_dict.keys())
        gene_list_len = len(gene_list)

        # gene_list_split = [gene_list[i:i + gene_list_len // self.thread] for i in range(0, gene_list_len, gene_list_len // self.thread)]
        gene_list_split = np.array_split(gene_list, self.thread)
        gene_list_split = [list(split) for split in gene_list_split]

        args = [(gene_list_split[i], self.mrna_dict, self.output) for i in range(self.thread)]

        # run the format_rreads_density function with multi-thread
        from multiprocessing import Pool
        pool = Pool(processes=self.thread)
        results = pool.map(self.format_reads, args)
        pool.close()
        pool.join()

        # merge the long data from each sub-list
        self.total_reads_df = pd.concat(results, ignore_index=True)
        
        numeric_cols = self.total_reads_df.select_dtypes(include='number').columns
        self.total_reads_df[numeric_cols] = self.total_reads_df[numeric_cols].astype(int)

    def output_density(self):
        
        # total_reads_df["sum"] = total_readsf_df[['frame_1', 'frame_2', 'frame_3']].sum(axis=1)
        # total_reads_df.to_csv(rna_attr.output + "_rna.txt", sep = '\t', index = False)
        total_reads_pl = pl.DataFrame(self.total_reads_df)
        total_reads_pl.write_csv(self.output + "_rna.txt", separator = '\t', has_header = True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project : riboParser
# @Script  : Density.py


import matplotlib
import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns
import numpy as np


class Density(object):

    def __init__(self, args):
        self.gene = args.list
        self.rpf = args.rpf
        self.sample_num = 0
        self.sample_name = None
        self.rel_name = None
        self.merged_rpf = None
        self.merged_rpf_na = None

        # for the high expression gene filter
        self.rpf_num = args.min
        self.high_gene = None
        self.high_rpf = None
        self.gene_rpf_sum = None
        self.total_rpf_num = None
        # filter the stable codon coverage range
        self.site = args.site
        self.frame = args.frame
        self.tis = args.tis
        self.tts = args.tts

        # table for codon occupancy
        self.occupancy = None
        self.occ_title = None
        self.rel_title = None
        self.codon_anno = {
            'AAA': ['Lys', 'K'], 'AAC': ['Asn', 'N'], 'AAG': ['Lys', 'K'], 'AAT': ['Asn', 'N'],
            'ACA': ['Thr', 'T'], 'ACC': ['Thr', 'T'], 'ACG': ['Thr', 'T'], 'ACT': ['Thr', 'T'],
            'AGA': ['Arg', 'R'], 'AGC': ['Ser', 'S'], 'AGG': ['Arg', 'R'], 'AGT': ['Ser', 'S'],
            'ATA': ['Ile', 'I'], 'ATC': ['Ile', 'I'], 'ATG': ['Met', 'M'], 'ATT': ['Ile', 'I'],
            'CAA': ['Gln', 'Q'], 'CAC': ['HIS', 'H'], 'CAG': ['Gln', 'Q'], 'CAT': ['HIS', 'H'],
            'CCA': ['Pro', 'P'], 'CCC': ['Pro', 'P'], 'CCG': ['Pro', 'P'], 'CCT': ['Pro', 'P'],
            'CGA': ['Arg', 'R'], 'CGC': ['Arg', 'R'], 'CGG': ['Arg', 'R'], 'CGT': ['Arg', 'R'],
            'CTA': ['Leu', 'L'], 'CTC': ['Leu', 'L'], 'CTG': ['Leu', 'L'], 'CTT': ['Leu', 'L'],
            'GAA': ['Glu', 'E'], 'GAC': ['Asp', 'D'], 'GAG': ['Glu', 'E'], 'GAT': ['Asp', 'D'],
            'GCA': ['Ala', 'A'], 'GCC': ['Ala', 'A'], 'GCG': ['Ala', 'A'], 'GCT': ['Ala', 'A'],
            'GGA': ['Gly', 'G'], 'GGC': ['Gly', 'G'], 'GGG': ['Gly', 'G'], 'GGT': ['Gly', 'G'],
            'GTA': ['Val', 'V'], 'GTC': ['Val', 'V'], 'GTG': ['Val', 'V'], 'GTT': ['Val', 'V'],
            'TAC': ['Tyr', 'Y'], 'TAT': ['Tyr', 'Y'], 'TCA': ['Ser', 'S'], 'TCC': ['Ser', 'S'],
            'TCG': ['Ser', 'S'], 'TCT': ['Ser', 'S'], 'TGC': ['Cys', 'C'], 'TGG': ['Trp', 'W'],
            'TGT': ['Cys', 'C'], 'TTA': ['Leu', 'L'], 'TTC': ['Phe', 'F'], 'TTG': ['Leu', 'L'],
            'TTT': ['Phe', 'F']
        }
        # except, 'TAA': ['*', '*'], 'TAG': ['*', '*'], 'TGA': ['*', '*']
        self.codon_table = None
        self.output = args.output

    def get_frame_rpf(self, rpf):
        if self.frame == 'all':
            for now_sp in self.sample_name:
                print('Import the RPFs file {file_name}.'.format(file_name=now_sp), flush=True)
                now_sp_index = [now_sp + '_f0', now_sp + '_f1', now_sp + '_f2']
                self.merged_rpf[now_sp] = rpf.loc[:, now_sp_index].sum(axis=1)
        elif self.frame == '0':
            for now_sp in self.sample_name:
                print('Import the RPFs file {file_name}.'.format(file_name=now_sp), flush=True)
                now_sp_index = [now_sp + '_f0']
                self.merged_rpf[now_sp] = rpf.loc[:, now_sp_index]
        elif self.frame == '1':
            for now_sp in self.sample_name:
                print('Import the RPFs file {file_name}.'.format(file_name=now_sp), flush=True)
                now_sp_index = [now_sp + '_f1']
                self.merged_rpf[now_sp] = rpf.loc[:, now_sp_index]
        elif self.frame == '2':
            for now_sp in self.sample_name:
                print('Import the RPFs file {file_name}.'.format(file_name=now_sp), flush=True)
                now_sp_index = [now_sp + '_f2']
                self.merged_rpf[now_sp] = rpf.loc[:, now_sp_index]

    def read_rpf(self):
        rpf = pd.read_csv(self.rpf, sep='\t', header=0, names=None)
        self.sample_num = int((len(rpf.columns) - 6) / 3)
        self.sample_name = pd.Series(rpf.columns[6::]).str[:-3].drop_duplicates().tolist()
        # merge the RPFs of nucleotide to codon
        # front 6 columns ['name', 'now_nt', 'from_tis', 'from_tes', 'region', 'codon']
        self.merged_rpf = rpf.iloc[:, 0:6].copy()
        self.get_frame_rpf(rpf)

        self.merged_rpf = self.merged_rpf[(self.merged_rpf["from_tis"] >= self.tis) &
                                          (self.merged_rpf["from_tts"] <= -self.tts)]
        # filter the specific genes
        if self.gene:
            gene_csv = pd.read_csv(self.gene, header=None, names=None)
            gene_name = []
            for gene in gene_csv.values.tolist():
                gene_name.extend(gene)
            self.merged_rpf = self.merged_rpf.loc[self.merged_rpf['name'].isin(gene_name)]
        else:
            pass

    def codon_occupancy(self):
        # filter the high expression gene
        self.total_rpf_num = self.merged_rpf[self.sample_name].sum()
        self.gene_rpf_sum = self.merged_rpf.groupby('name')[self.sample_name].sum()
        self.high_gene = self.gene_rpf_sum.loc[self.gene_rpf_sum.max(axis=1) >= self.rpf_num, :].index
        self.high_rpf = self.merged_rpf.loc[self.merged_rpf['name'].isin(self.high_gene), :]

        # calc the average density of each gene
        ave_rpf = self.high_rpf.replace(0, np.nan).groupby('name')[self.sample_name].mean()
        ave_dst_title = [i + '_ave_dst' for i in self.sample_name]
        np.seterr(divide='ignore', invalid='ignore')
        # ave_dst = self.high_rpf.groupby('name').apply(lambda x: np.divide(x[self.sample_name], ave_rpf.loc[x.name]))
        # ave_dst.columns = ave_dst_title
        # high_rpf_ave_dst = pd.concat([self.high_rpf, ave_dst], axis=1)
        # self.occupancy = high_rpf_ave_dst.replace(0, np.nan).groupby('codon')[ave_dst_title].mean()

        # codon shift by E/P/A site
        if self.site == 'E':
            shift_codon = -1
        elif self.site == 'P':
            shift_codon = 0
        elif self.site == 'A':
            shift_codon = 1
        else:
            shift_codon = 0

        # calculate the average density of each codon
        high_rpf_group = self.high_rpf.groupby('name')
        now_num = 0
        ave_dst_list = []
        for gene, rpfs in high_rpf_group:
            # counter
            now_num += 1
            if now_num % 100 == 0:
                print('Row: {rows}, {gene}.'.format(rows=now_num, gene=gene), flush=True)
            if now_num == len(high_rpf_group):
                print('Row: {rows}, {gene}.'.format(rows=now_num, gene=gene), flush=True)

            # prepare the gene df
            temp_rpfs = rpfs.loc[:, self.sample_name].copy().shift(shift_codon).fillna(0)
            temp_ave_dst = np.divide(temp_rpfs, ave_rpf.loc[gene])
            ave_dst_list.append(temp_ave_dst)

        # merge average density and calculate occupancy
        ave_dst = pd.concat(ave_dst_list, axis=0)
        ave_dst.columns = ave_dst_title
        high_rpf_ave_dst = pd.concat([self.high_rpf, ave_dst], axis=1)
        self.occupancy = high_rpf_ave_dst.replace(0, np.nan).groupby('codon')[ave_dst_title].mean()

        for codon in self.codon_anno.keys():
            # merge the codon count
            try:
                self.codon_anno[codon].extend(self.occupancy.loc[codon].values.tolist())
            except KeyError:
                self.codon_anno[codon].extend([np.nan]*self.sample_num)

        self.codon_table = pd.DataFrame(self.codon_anno).T

        # z-score standard
        rel_occ = (self.codon_table.iloc[:, 2::] -
                   self.codon_table.iloc[:, 2::].mean())/self.codon_table.iloc[:, 2::].std()
        self.codon_table = pd.concat([self.codon_table, rel_occ], axis=1)
        self.occ_title = [i + '_occupancy' for i in self.sample_name]
        self.rel_title = [i + '_relative' for i in self.sample_name]
        self.codon_table.columns = ['AA', 'Abbr.'] + self.occ_title + self.rel_title
        self.codon_table.sort_values('Abbr.', inplace=True)
        # self.codon_table = self.codon_table.round(2)
        self.codon_table.index.name = 'codon'
        self.codon_table.to_csv(self.output + '_codon_occupancy.txt', header=True, index=True, sep='\t')

    def draw_occupancy_corr(self):
        out_pdf = self.output + "_occupancy_corrplot.pdf"
        out_png = self.output + "_occupancy_corrplot.png"

        occ_corr = self.codon_table[self.occ_title].astype('float').corr()
        occ_corr.to_csv(self.output + '_occupancy_corr.txt', sep='\t', index=True)

        matplotlib.use('AGG')
        sns.heatmap(data=occ_corr, annot=True, fmt=".4f", cmap="vlag", linewidths=.01,
                    cbar_kws={"orientation": "horizontal"})
        plt.tight_layout()
        plt.show()

        plt.savefig(fname=out_pdf)
        plt.savefig(fname=out_png)
        plt.close()

    def draw_occupancy_heat(self):
        # output the absolute occupancy heatmap
        out_pdf = self.output + "_occupancy_heatplot.pdf"
        out_png = self.output + "_occupancy_heatplot.png"

        matplotlib.use('AGG')
        fig = plt.figure(figsize=(6 + 0.2 * self.sample_num, 20), dpi=300)
        sns.heatmap(data=self.codon_table[self.occ_title].astype('float'),
                    annot=True, fmt=".4f", cmap="vlag", linewidths=.01,
                    cbar_kws={"orientation": "horizontal", "pad": 0.05})
        plt.tight_layout()
        plt.show()

        plt.savefig(fname=out_pdf)
        plt.savefig(fname=out_png)
        plt.close()

    def draw_occupancy_relative_heat(self):

        # output the relative occupancy heatmap
        out_pdf = self.output + "_occupancy_relative_heatplot.pdf"
        out_png = self.output + "_occupancy_relative_heatplot.png"

        matplotlib.use('AGG')
        fig = plt.figure(figsize=(6 + 0.2 * self.sample_num, 20), dpi=300)
        sns.heatmap(data=self.codon_table[self.rel_title].astype('float'),
                    annot=True, fmt=".4f", cmap="vlag", linewidths=.01,
                    cbar_kws={"orientation": "horizontal", "pad": 0.05})
        plt.tight_layout()
        plt.show()

        plt.savefig(fname=out_pdf)
        plt.savefig(fname=out_png)
        plt.close()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project : riboParser
# @Script  : Codon.py


import CAI
import matplotlib
import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns
import numpy as np
from . import RPFs


class Codon(object):

    def __init__(self, args):
        # input and output
        self.gene = args.list
        self.rpf = args.rpf
        self.output = args.output

        # for the high expression gene filter
        self.sample_num = 0
        self.sample_name = None
        self.rel_name = None
        self.merged_rpf = None

        self.rpf_num = args.min
        self.high_gene = None
        self.high_rpf = None
        self.gene_rpf_sum = None
        self.total_rpf_num = None
        self.total_rpf_num = None

        # filter the stable codon coverage range
        self.tis = args.tis
        self.tts = args.tts
        self.site = args.site
        self.frame = args.frame
        # self.mrna_seq = args.sequence
        # self.mrna_dict = OrderedDict()
        # self.longest = str(args.longest)
        self.codon_dict, self.codon_table = RPFs.codon_table()
        self.codon_title = ['AA', 'Abbr.']

    def read_rpf(self):
        # raw_rpf, sample_name, sample_num, merged_rpf, total_rpf_num, gene_rpf_sum, high_gene, high_rpf
        rpf_results = RPFs.import_rpf(rpf_file=self.rpf, sites=self.site, frame=self.frame,
                                      # tis=self.tis, tts=self.tts,
                                      sample_num=None, sample_name=None,
                                      gene=self.gene, rpf_num=self.rpf_num)
        # self.raw_rpf = rpf_results[0]
        self.sample_name = rpf_results[1]
        self.sample_num = rpf_results[2]
        self.merged_rpf = rpf_results[3]
        self.total_rpf_num = rpf_results[4]
        self.gene_rpf_sum = rpf_results[5]
        # self.high_gene = rpf_results[6]
        self.high_rpf = rpf_results[7]
        self.high_rpf[self.sample_name] = self.high_rpf[self.sample_name] * 10000000 / self.total_rpf_num

    def merge_codon_table(self, count, freq, rscu):
        # synonymous codon


        # merge all codon usage
        for codon in self.codon_dict.keys():
            # merge the codon count
            try:
                self.codon_dict[codon].append(round(count[codon], 3))
            except KeyError:
                self.codon_dict[codon].append(np.nan)
            # merge the frequency of the codon counts of per 1000
            try:
                self.codon_dict[codon].append(round(freq[codon], 3))
            except KeyError:
                self.codon_dict[codon].append(np.nan)
            # merge the RSCU
            try:
                self.codon_dict[codon].append(round(rscu[codon], 3))
            except KeyError:
                self.codon_dict[codon].append(np.nan)

    def get_all_codon_usage(self):
        print('Whole gene sequence.', flush=True)
        all_cds = self.merged_rpf.loc[self.merged_rpf['region'] == 'cds', 'codon']
        all_cds_codon_count = all_cds.value_counts()
        all_codon_freq = round(all_cds_codon_count * 1000 / all_cds_codon_count.sum(), 3)
        all_cds_codon_count_dict = all_cds_codon_count.to_dict()
        all_codon_freq_dict = all_codon_freq.to_dict()
        all_rscu_dict = CAI.RSCU(all_cds.to_list())
        self.merge_codon_table(all_cds_codon_count_dict, all_codon_freq_dict, all_rscu_dict)
        self.codon_title.extend(['All_Count', 'All_Freq', 'All_RSCU'])

    def get_high_codon_usage(self, high_rpf_cds, sample):
        cds = high_rpf_cds
        cds_codon_count = cds.value_counts()
        codon_freq = round(cds_codon_count * 1000 / cds_codon_count.sum(), 3)
        cds_codon_count_dict = cds_codon_count.to_dict()
        codon_freq_dict = codon_freq.to_dict()
        rscu_dict = CAI.RSCU(cds.to_list())
        self.merge_codon_table(cds_codon_count_dict, codon_freq_dict, rscu_dict)
        self.codon_title.extend([sample + '_Count', sample + '_Freq', sample + '_RSCU'])

    def get_codon_usage(self):

        for sample in self.sample_name:
            print('Now sample is: {sample}.'.format(sample=sample), flush=True)
            # filter the high expression genes
            high_gene = self.gene_rpf_sum[self.gene_rpf_sum[sample] > self.rpf_num]
            gene_ids = list(high_gene.index)
            high_rpf = self.merged_rpf[self.merged_rpf['name'].isin(gene_ids)].copy().reset_index(drop=True)
            high_rpf_cds = high_rpf.loc[high_rpf['region'] == 'cds', 'codon']
            self.get_high_codon_usage(high_rpf_cds, sample)

        self.codon_table = pd.DataFrame(self.codon_dict).T
        self.codon_table.columns = self.codon_title
        self.codon_table.sort_values('Abbr.', inplace=True)
        self.codon_table.index.name = 'codon'
        self.codon_table = self.codon_table.reset_index(drop=False)

        self.codon_table.to_csv(self.output + '_codon_usage.txt', header=True, index=False, sep='\t')

    def draw_whole_gene(self, codon_table):
        counts = codon_table['All_Count']
        freq = codon_table['All_Freq']
        rscu = codon_table['All_RSCU']

        # draw the figure
        matplotlib.use('AGG')

        # output figure name
        out_pdf = self.output + "_codon_usage.pdf"
        out_png = self.output + "_codon_usage.png"
        fig = plt.figure(figsize=(12, 6), dpi=300)

        ax1 = fig.add_subplot(311)
        sns.barplot(x=np.arange(0, 61), y=counts, ax=ax1)
        ax1.set_title('codon counts')
        ax1.set_xticks(np.arange(0, 61))
        ax1.set_xticklabels(codon_table['xticks'], rotation=90)

        ax2 = fig.add_subplot(312)
        sns.barplot(x=np.arange(0, 61), y=freq, ax=ax2)
        ax2.set_title('codon frequency')
        ax2.set_xticks(np.arange(0, 61))
        ax2.set_xticklabels(codon_table['xticks'], rotation=90)

        ax3 = fig.add_subplot(313)
        sns.barplot(x=np.arange(0, 61), y=rscu, ax=ax3)
        ax3.set_title('RSCU')
        ax3.set_xticks(np.arange(0, 61))
        ax3.set_xticklabels(codon_table['xticks'], rotation=90)

        plt.tight_layout()
        plt.show()
        plt.close()
        fig.savefig(fname=out_pdf)
        fig.savefig(fname=out_png)

    def draw_codon_usage(self):
        codon_table = self.codon_table.copy()
        codon_table['codon'] = codon_table.index
        codon_table.sort_values(by=['Abbr.', 'codon'], ascending=[True, True], inplace=True)
        codon_table['xticks'] = codon_table.index + '[' + codon_table['Abbr.'] + ']'

        self.draw_whole_gene(codon_table)

        for num in range(self.sample_num):
            counts = codon_table[self.sample_name[num] + '_Count']
            freq = codon_table[self.sample_name[num] + '_Freq']
            rscu = codon_table[self.sample_name[num] + '_RSCU']

            # draw the figure
            matplotlib.use('AGG')

            # output figure name
            out_pdf = self.sample_name[num] + "_codon_usage.pdf"
            out_png = self.sample_name[num] + "_codon_usage.png"
            fig = plt.figure(figsize=(12, 6), dpi=300)

            ax1 = fig.add_subplot(311)
            sns.barplot(x=np.arange(0, 61), y=counts, ax=ax1)
            ax1.set_title('codon counts')
            ax1.set_xticks(np.arange(0, 61))
            ax1.set_xticklabels(codon_table['xticks'], rotation=90)

            ax2 = fig.add_subplot(312)
            sns.barplot(x=np.arange(0, 61), y=freq, ax=ax2)
            ax2.set_title('codon frequency')
            ax2.set_xticks(np.arange(0, 61))
            ax2.set_xticklabels(codon_table['xticks'], rotation=90)

            ax3 = fig.add_subplot(313)
            sns.barplot(x=np.arange(0, 61), y=rscu, ax=ax3)
            ax3.set_title('RSCU')
            ax3.set_xticks(np.arange(0, 61))
            ax3.set_xticklabels(codon_table['xticks'], rotation=90)

            plt.tight_layout()
            plt.show()
            plt.close()
            fig.savefig(fname=out_pdf)
            fig.savefig(fname=out_png)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project : riboParser
# @Script  : rpf_digest.py


from .ribo.ArgsParser import *
from .ribo.Digestion import *


def main():
    print('\nDetect the digestion sites.\n', flush=True)
    print('\nStep1: Checking the input Arguments.\n', flush=True)
    args = digestion_args_parser()
    ribo_attr = Ribo(args)

    print('\nStep2: Import the annotation of transcripts.\n', flush=True)
    ribo_attr.read_transcript()

    print('\nStep3: Detect the digestion sites.\n', flush=True)
    ribo_attr.get_digest_sites()

    print('\nStep4: Output the digestion sites.\n', flush=True)
    ribo_attr.output_digest_sites()

    print('\nStep5: Draw the heatmap of digestion sites.\n', flush=True)
    ribo_attr.digestion_plot()

    print('\nStep6: Draw the seq logo of digestion sites.\n', flush=True)
    ribo_attr.output_counts()
    
    ribo_attr.seq_logo_plot2()

    print('\nAll done.\n', flush=True)
    now_time()


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project : riboParser
# @Script  : Quant.py


import matplotlib.pyplot as plt
import seaborn as sns

import plotly.graph_objs as go
import plotly.express as px
from scipy.spatial.distance import pdist
from scipy.cluster import hierarchy

import numpy as np
import pandas as pd
from . import RPFs


class Quant(object):
    def __init__(self, args):
        self.rpf_file = args.rpf

        # self.anno = args.anno
        self.output_prefix = args.output

        # parameters for the trimmed CDS region
        self.rpf_num = 0
        self.site = 'P'
        self.tis = args.tis
        self.tts = args.tts
        self.frame = args.frame
        self.length = None

        # parameters for rpf file parser
        self.sample_name = None
        self.sample_num = None
        self.merged_rpf = None
        self.total_rpf = None
        self.col_index = None

        # parameters for different region
        self.utr5 = args.utr5
        self.utr3 = args.utr3

        self.utr5_rpf = None
        self.utr5_rpm = None
        self.utr5_tpm = None

        self.utr3_rpf = None
        self.utr3_rpm = None
        self.utr3_tpm = None

        self.cds_rpf = None
        self.cds_rpm = None
        self.cds_tpm = None

    def read_rpf(self):
        # raw_rpf, sample_name, sample_num, merged_rpf, total_rpf_num, gene_rpf_sum, high_gene, high_rpf
        rpf_results = RPFs.import_rpf(rpf_file=self.rpf_file,
                                      sites=self.site,
                                      frame=self.frame,
                                      sample_num=None,
                                      sample_name=None,
                                      tis=None,
                                      tts=None,
                                      gene=None,
                                      rpf_num=self.rpf_num)
        # self.raw_rpf = rpf_results[0]
        self.sample_name = rpf_results[1].copy()
        # self.sample_num = rpf_results[2]
        self.merged_rpf = rpf_results[3].copy()
        self.total_rpf = rpf_results[4]
        # self.gene_rpf_sum = rpf_results[5]
        # self.high_gene = rpf_results[6]
        # self.high_rpf = rpf_results[7].copy()
        del rpf_results

        self.sample_num = len(self.sample_name)
        self.col_index = ['name'] + self.sample_name
        self.length = self.merged_rpf.loc[:, ['name', 'region']].groupby("name").count() * 3

    def output_total_rpf(self):
        total_rpf_name = self.output_prefix + "_total.txt"
        self.total_rpf.index.name = 'sample'
        self.total_rpf.name = 'rpf_count'

        self.total_rpf.to_csv(total_rpf_name, sep='\t', index=True)

    def output_gene_rpf(self, rpf_count, rpf_rpm, rpf_rpkm, rpf_tpm, region):

        # output the rpf / rpm / tpm results
        rpf_count.astype(int)
        rpf_count.index.name = "name"
        rpf_count.to_csv(self.output_prefix + "_" + region + "_rpf_quant.txt", sep='\t', index=True)

        rpf_rpm = rpf_rpm.astype(float).round(2)
        rpf_rpm.index.name = "name"
        rpf_rpm.to_csv(self.output_prefix + "_" + region + "_rpm_quant.txt", sep='\t', index=True)

        rpf_rpkm = rpf_rpkm.astype(float).round(2)
        rpf_rpkm.index.name = "name"
        rpf_rpkm.to_csv(self.output_prefix + "_" + region + "_rpkm_quant.txt", sep='\t', index=True)

        rpf_tpm = rpf_tpm.astype(float).round(2)
        rpf_tpm.index.name = "name"
        rpf_tpm.to_csv(self.output_prefix + "_" + region + "_tpm_quant.txt", sep='\t', index=True)

    def quant_utr5_rpf(self, region='5utr'):

        print("Summary RPFs of 5'-UTR.", flush=True)
        # filter the gene region
        self.utr5_rpf = self.merged_rpf.loc[self.merged_rpf["region"] == region, self.col_index]
        self.utr5_rpf = self.utr5_rpf.groupby("name")[self.sample_name].sum()

        self.utr5_rpf.columns = [i + "_" + region + "_rpf" for i in self.sample_name]

        # calculate the gene rpm
        self.utr5_rpm = np.divide(self.utr5_rpf, np.asarray(self.total_rpf)) * 1e6
        self.utr5_rpm.columns = [i + "_" + region + "_rpm" for i in self.sample_name]

        # calculate the gene rpkm
        self.utr5_rpkm = (self.utr5_rpf * 1e9).div(self.length['region'], axis=0).div(np.asarray(self.total_rpf), axis=1)
        self.utr5_rpkm.columns = [i + "_" + region + "_rpkm" for i in self.sample_name]

        # calculate the gene tpm
        utr5_rpk = (self.utr5_rpf * 1000).div(self.length['region'], axis=0).dropna()
        utr5_rpk_sum = utr5_rpk.sum(axis=0)
        self.utr5_tpm = (utr5_rpk * 1000000).div(utr5_rpk_sum, axis=1)
        self.utr5_tpm.columns = [i + "_" + region + "_tpm" for i in self.sample_name]

        self.output_gene_rpf(self.utr5_rpf, self.utr5_rpm, self.utr5_rpkm, self.utr5_tpm, 'utr5')

    def quant_utr3_rpf(self, region='3utr'):

        print("Summary RPFs of 3'-UTR.", flush=True)
        # filter the gene region
        self.utr3_rpf = self.merged_rpf.loc[self.merged_rpf["region"] == region, self.col_index]
        self.utr3_rpf = self.utr3_rpf.groupby("name")[self.sample_name].sum()

        self.utr3_rpf.columns = [i + "_" + region + "_rpf" for i in self.sample_name]

        # calculate the gene rpm
        self.quant_rpm = np.divide(self.utr3_rpf, np.asarray(self.total_rpf)) * 1e6
        self.quant_rpm.columns = [i + "_" + region + "_rpm" for i in self.sample_name]

        # calculate the gene rpkm
        self.utr3_rpkm = (self.utr3_rpf * 1e9).div(self.length['region'], axis=0).div(np.asarray(self.total_rpf), axis=1)
        self.utr3_rpkm.columns = [i + "_" + region + "_rpkm" for i in self.sample_name]

        # calculate the gene tpm
        utr3_rpk = (self.utr3_rpf * 1000).div(self.length['region'], axis=0).dropna()
        utr3_rpk_sum = utr3_rpk.sum(axis=0)
        self.utr3_tpm = (utr3_rpk * 1000000).div(utr3_rpk_sum, axis=1)
        self.utr3_tpm.columns = [i + "_" + region + "_tpm" for i in self.sample_name]

        self.output_gene_rpf(self.utr3_rpf, self.utr3_rpm, self.utr3_rpkm, self.utr3_tpm, 'utr3')

    def quant_cds_rpf(self, region='cds'):

        print('Summary RPFs of CDS.', flush=True)
        # filter the gene region
        self.cds_rpf = self.merged_rpf[self.merged_rpf['from_tis'] >= self.tis]
        self.cds_rpf = self.cds_rpf[self.cds_rpf['from_tts'] < -self.tts]

        self.cds_rpf = self.cds_rpf.loc[self.cds_rpf["region"] == region, self.col_index]
        self.cds_rpf = self.cds_rpf.groupby("name")[self.sample_name].sum()

        self.cds_rpf.columns = [i + "_" + region + "_rpf" for i in self.sample_name]

        # calculate the gene rpm
        self.cds_rpm = np.divide(self.cds_rpf, np.asarray(self.total_rpf)) * 1e6
        self.cds_rpm.columns = [i + "_" + region + "_rpm" for i in self.sample_name]

        # calculate the gene rpkm
        # formula: RPKM = (10^9 * C) / (N * L)
        # C: number of reads mapped to a gene
        # N: total number of reads mapped to the reference
        # L: number of bases in a gene
        
        self.cds_rpkm = (self.cds_rpf * 1e9).div(self.length['region'], axis=0).div(np.asarray(self.total_rpf), axis=1)
        self.cds_rpkm.columns = [i + "_" + region + "_rpkm" for i in self.sample_name]

        # calculate the gene tpm
        cds_rpk = (self.cds_rpf * 1000).div(self.length['region'], axis=0).dropna()
        cds_rpk_sum = cds_rpk.sum(axis=0)
        self.cds_tpm = (cds_rpk * 1000000).div(cds_rpk_sum, axis=1)
        self.cds_tpm.columns = [i + "_" + region + "_tpm" for i in self.sample_name]

        self.output_gene_rpf(self.cds_rpf, self.cds_rpm, self.cds_rpkm, self.cds_tpm, 'cds')

    def quant_region(self):
        if self.utr3:
            self.quant_utr5_rpf(region='utr5')

        self.quant_cds_rpf(region='cds')

        if self.utr5:
            self.quant_utr3_rpf(region='utr3')

    def get_sub_plot_num(self):
        factor1 = int(self.sample_num ** 0.5)
        factor2 = int(self.sample_num ** 0.5)
        square = factor1 * factor2

        while not square >= self.sample_num:
            if factor2 > factor1:
                factor1 += 1
            else:
                factor2 += 1
            square = factor1 * factor2

        return factor1, factor2

    def draw_rpf_barplot(self):
        """
        draw the rpf region of each samples with plotly

        1. summary the rpf num of each region [5'-utr, cds, 3'-utr]
        2. draw the stacked bar plot
        """
        
        out_pdf = self.output_prefix + "_rpf_barplot.pdf"
        # out_png = self.output_prefix + "_rpf_barplot.png"

        rpf_sum = self.merged_rpf.groupby('region')[self.sample_name].apply(sum)
        rpf_sum_per = round(rpf_sum.div(rpf_sum.sum()) * 100, 2)

        rpf_sum_per_t = rpf_sum_per.reset_index().melt(id_vars="region",
                                                       var_name="samples",
                                                       value_name="proportion")

        fig = px.bar(rpf_sum_per_t,
                     x="samples",
                     y="proportion",
                     color="region",
                     orientation="v",
                     text_auto=True,
                     category_orders={"region": ["5utr", "cds", "3utr"]},
                     color_discrete_sequence=px.colors.qualitative.Pastel2)

        fig.update_layout(xaxis_title="proportion (%)",
                          yaxis_title="samples",
                          template="simple_white",
                          height=500, width=300 + len(self.sample_name) * 20)

        fig.write_image(out_pdf, engine="kaleido")
        # fig.write_image(out_png, engine="kaleido")

    def draw_rpf_cdfplot(self):
        """
        draw the cdf figure of each samples with plotly
        to check the global expression levels

        1. log the cds rpm
        2. draw the eCDF plot of each sample with plotly
        """
        out_pdf = self.output_prefix + "_rpf_cdfplot.pdf"
        # out_png = self.output_prefix + "_rpf_cdfplot.png"

        # get the log2 values
        cds_log = np.log2(self.cds_rpm + 1)
        cds_log = cds_log.reset_index().melt(id_vars="name", var_name="samples", value_name="rpm")

        # draw the proportion of RPFs in different region
        fig = px.ecdf(cds_log, x="rpm", color="samples")

        fig.update_layout(xaxis_title="eCDF",
                          yaxis_title="log2 RPM",
                          template="simple_white",
                          height=600, width=650)

        fig.write_image(out_pdf, engine="kaleido")
        # fig.write_image(out_png, engine="kaleido")

    def draw_rpf_pcaplot(self):
        """
        draw the pca figure of each samples with plotly

        1. log the cds rpm
        2. calculate the components of each dimension
        3. draw the PCA scatter plot of each sample
        """
        from sklearn.decomposition import PCA

        out_pdf = self.output_prefix + "_rpf_pcaplot.pdf"
        # out_png = self.output_prefix + "_rpf_pcaplot.png"

        # run the PCA with sklearn function
        out_txt = self.output_prefix + "_cds_rpf_pca.txt"

        pca = PCA(n_components=2)
        pca.fit(np.log2(self.cds_rpm.T + 1))
        variance_ratios = pca.explained_variance_ratio_
        # covariance = pca.get_covariance() * 100
        components = pca.fit_transform(np.log2(self.cds_rpm.T + 1))

        pca_df = pd.DataFrame(data=components, columns=['PC1', 'PC2'])
        pca_df.index = self.sample_name
        pca_df = pca_df.reset_index().rename(columns={'index': 'samples'})

        pca_df.to_csv(out_txt, sep='\t', index=True)

        pca_df_t = pca_df.reset_index().melt(id_vars="index",
                                             var_name="pc",
                                             value_name="value").rename(columns={'index': 'samples'})

        # draw the PCA figure
        fig = px.scatter(pca_df, x="PC1", y="PC2", color="samples")

        fig.update_xaxes(tickformat='.2e')
        fig.update_yaxes(tickformat='.2e')
        fig.update_layout(xaxis_title="PC1 (" + str("{:.2f}".format(variance_ratios[0])) + "%)",
                          yaxis_title="PC2 (" + str("{:.2f}".format(variance_ratios[1])) + "%)",
                          template="simple_white",
                          height=400, width=490)

        fig.write_image(out_pdf, engine="kaleido")

    def draw_rpf_heatmap1(self):
        """
        draw the gene rpm heatmap of each samples

        1. log the cds rpm
        2. scale the rpm to center 0
        3. draw the heatmap with di-cluster
        """

        out_pdf = self.output_prefix + "_rpf_heatmap.pdf"
        out_png = self.output_prefix + "_rpf_heatmap.png"

        cds_rpm = self.cds_rpm[(self.cds_rpm > 0).any(axis=1)]
        rpm_log = np.log2(cds_rpm + 1)

        plt.figure(figsize=(8, 12), dpi=300)

        sns.clustermap(rpm_log,
                       cmap="RdBu",
                       center=0,
                       dendrogram_ratio=(.1, .2),
                       cbar_pos=(.02, .02, .025, .17))

        plt.savefig(fname=out_pdf)
        plt.savefig(fname=out_png)

        plt.close()

    def draw_rpf_heatmap2(self):
        """
        draw the gene rpm heatmap of each samples with plotly

        1. log the cds rpm
        2. calculate the distance of rpm data matrix and clustered with scipy
        3. draw the di-cluster heatmap with plotly
        """

        out_pdf = self.output_prefix + "_rpf_heatmap.pdf"
        # out_png = self.output_prefix + "_rpf_heatmap.png"

        rpm_log = np.log2(self.cds_rpm[(self.cds_rpm > 0).any(axis=1)] + 1)

        cds_dist = pdist(rpm_log)
        clustering = hierarchy.linkage(cds_dist, method='ward', metric='euclidean')

        row_idx = hierarchy.leaves_list(clustering)
        rpm_clustered = rpm_log.iloc[row_idx, :]

        fig = go.Figure(
            data=go.Heatmap(x=rpm_clustered.columns.values,
                            y=rpm_clustered.index.values,
                            z=rpm_clustered,
                            # zmin=-3, zmax=3,
                            colorscale='RdBu',
                            reversescale=True)
        )

        fig.update_layout(xaxis=dict(side='bottom'),
                          yaxis=dict(side='left'),
                          height=600, width=450,
                          title=dict(text='Expression pattern', x=0.5, xanchor='center'),
                          xaxis_showgrid=False, yaxis_showgrid=False,
                          xaxis_tickangle=-90, yaxis_tickangle=0)

        fig.write_image(out_pdf, engine="kaleido")
        # fig.write_image(out_png, engine="kaleido")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project : riboParser
# @Script  : rpf_coverage.py


from .ribo import ArgsParser
from .ribo import Coverage


def main():
    ArgsParser.now_time()
    print('\nDraw the metagene coverage.\n', flush=True)
    print('Step1: Checking the input Arguments.', flush=True)
    args = ArgsParser.coverage_args_parser()
    meta = Coverage.Coverage(args)

    print('\nStep2: Import the RPFs file.', flush=True)
    meta.read_rpf()
    meta.import_gene()

    print('\nStep3: Adjust mRNAs to the same dimension.', flush=True)
    meta.process_utr5()
    meta.process_cds()
    meta.process_utr3()

    print('\nStep5: Draw the line plot of metagene coverage.', flush=True)
    meta.draw_meta_gene_line()

    print('\nStep6: Draw the heatmap of metagene coverage.', flush=True)
    if args.heatmap:
        meta.draw_meta_gene_heat()

    print('\nStep7: Output the metagene coverage.', flush=True)
    meta.output_meta_gene()

    print('\nAll done.\n', flush=True)
    ArgsParser.now_time()


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project : riboParser
# @Script  : rpf_end.py


import os
import sys
from collections import OrderedDict
from itertools import islice
from Bio import SeqIO
import argparse
import pandas as pd
import pysam


class Mrna(object):

    def __init__(self, record):
        self.chromosome = record[0]
        self.gene_id = record[1]
        self.gene_name = record[2]
        self.transcript_id = record[3]
        self.start = record[4]
        self.end = record[5]
        self.utr5_length = int(record[6])
        self.cds_length = int(record[7])
        self.utr3_length = int(record[8])
        self.strand = record[9]
        self.rep_transcript = record[10]
        self.modified = record[11]
        self.bam = []
        self.rpf = []
        self.seq = None


class Ribo(object):

    def __init__(self, args):
        self.mrna_file = args.transcript
        self.mrna_seq = args.sequence
        self.mrna_dict = OrderedDict()
        self.longest = str(args.longest)

        self.bam = args.bam
        self.sample_format = os.path.splitext(args.bam)[1]
        self.pysam_input = None
        self.pysam_output = None

        # self.psite = args.psite
        self.min_length = args.min
        self.max_length = args.max
        self.offset = {}
        self.profile = None

        self.output_prefix = args.output

    def read_transcript(self):

        record = SeqIO.parse(self.mrna_seq, "fasta")
        mrna_sequence = OrderedDict()
        for line in record:
            # sys.stdout.writelines("import gene:  {gene}\n".format(gene=line.id))
            mrna_sequence[line.id] = line.seq

        # mrna_sequence = pysam.FastaFile(self.mrna_seq)

        with open(self.mrna_file, 'r') as trans_file_in:
            if self.longest:
                for line in islice(trans_file_in, 1, None):
                    record = line.strip().split('\t')
                    # if int(record[7]) % 3 != 0:
                    #     sys.stdout.write("Delete {gene}. CDS length doesn't fit 3nt periodicity. \n".format(gene=record[3]))
                    #     continue

                    if record[10] == "True":
                        now_mrna = Mrna(record)
                        # sys.stdout.writelines(now_mrna.transcript_id + '\n')
                        now_mrna.length = now_mrna.utr5_length + now_mrna.utr3_length + now_mrna.cds_length
                        try:
                            now_mrna.seq = mrna_sequence[now_mrna.transcript_id]
                            now_mrna.rpf = [0] * len(now_mrna.seq)
                            self.mrna_dict[now_mrna.transcript_id] = now_mrna
                        except KeyError:
                            continue
                    else:
                        continue
            else:
                for line in islice(trans_file_in, 1, None):
                    record = line.strip().split('\t')
                    now_mrna = Mrna(record)
                    # sys.stdout.writelines(now_mrna.transcript_id + '\n')
                    now_mrna.length = now_mrna.utr5_length + now_mrna.utr3_length + now_mrna.cds_length
                    try:
                        now_mrna.seq = mrna_sequence[now_mrna.transcript_id]
                        now_mrna.rpf = [0] * len(now_mrna.seq)
                        self.mrna_dict[now_mrna.transcript_id] = now_mrna
                    except KeyError:
                        continue

    @staticmethod
    def get_tes_site(utr5, cds, utr3):
        start_site = (utr5 + cds) - 100
        end_site = (utr5 + cds) + 100
        if start_site < 0:
            start_site = 0
        if end_site > utr5 + cds + utr3:
            end_site = utr5 + cds + utr3

        return start_site, end_site

    @staticmethod
    def get_tis_site(utr5, cds, utr3):
        start_site = utr5 - 100
        end_site = utr5 + 100
        if start_site < 0:
            start_site = 0
        if end_site > utr5 + cds:
            end_site = utr5 + cds

        return start_site, end_site

    def rpf_end_sum(self):
        rpf_list = []
        self.pysam_input = pysam.AlignmentFile(self.bam, "rb")
        self.pysam_output = pysam.AlignmentFile(self.output_prefix + "_end.bam", 'wb', template=self.pysam_input)
        for name, isoform in self.mrna_dict.items():
            start_site, end_site = self.get_tis_site(isoform.utr5_length, isoform.cds_length, isoform.utr3_length)
            for read in self.pysam_input.fetch(name, start_site, end_site):
                self.pysam_output.write(read)

            start_site, end_site = self.get_tes_site(isoform.utr5_length, isoform.cds_length, isoform.utr3_length)
            for read in self.pysam_input.fetch(name, start_site, end_site):
                self.pysam_output.write(read)

                # if self.min_length <= read_length <= self.max_length:
                #     try:
                #         p_site = map_start + self.offset[read_length][0]
                #         self.mrna_dict[line.reference_name].rpf[p_site] += 1
                #         p_site = map_start + self.offset[read_length][1]
                #         self.mrna_dict[line.reference_name].rpf[p_site] += 1
                #     except KeyError:
                #         # sys.stdout.writelines("skip reads {reads}!".format(reads=line))
                #         continue
                #     except IndexError:
                #         continue
                # else:
                #     pass


def rpf_end_args_parser():
    parser = argparse.ArgumentParser(description="This script is used to parsing the ribo-seq.")

    # arguments for the Required arguments
    input_group = parser.add_argument_group('Required arguments')
    input_group.add_argument(
        '-t', dest="transcript", required=True, type=str,
        help="the name of input transcript file in TXT format."
    )
    input_group.add_argument(
        '-s', dest="sequence", required=True, type=str,
        help="the name of input transcript sequence file in FA format."
    )
    input_group.add_argument(
        '-b', dest="bam", required=True, type=str,
        help="the name of mapping file in BAM format."
    )
    input_group.add_argument(
        '-o', dest="output", required=True, type=str,
        help="the prefix of output file. (prefix + _rpf.txt)"
    )

    # arguments for the ribo-seq parsing
    parser.add_argument(
        '-l', dest="longest", action='store_true', required=False, default=False,
        help="only retain the transcript with longest CDS of each gene (default: %(default)s)."
             "Recommended : True"
    )
    parser.add_argument(
        '-m', dest="min", required=False, type=int, default=27,
        help="the minimum reads length to keep (default: %(default)s nt)."
    )
    parser.add_argument(
        '-M', dest="max", required=False, type=int, default=33,
        help="the maximum reads length to keep (default: %(default)s nt)."
    )

    args = parser.parse_args()
    return args


def main():

    sys.stdout.writelines('\nDetect the p-site offset.\n')
    sys.stdout.writelines('Step1: Checking the input Arguments.\n')
    args = rpf_end_args_parser()
    ribo_attr = Ribo(args)

    sys.stdout.writelines('\nStep2: Import the transcripts annotation.\n')
    ribo_attr.read_transcript()

    sys.stdout.writelines('\nStep3: Import the bam file.\n')
    ribo_attr.rpf_end_sum()

    sys.stdout.writelines('\nAll done.\n')


if __name__ == '__main__':
    main()




#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# @Project : riboParser
# @Script  : merge_dst_list.py


import os
import pandas as pd
import argparse


def merge_density_args_parser():

    parser = argparse.ArgumentParser(description='This script is used to create the list of density files')

    # needed arguments
    input_group = parser.add_argument_group('Required arguments')
    input_group.add_argument(
        "-l", "--list", nargs='+', required=True, help="List for density files (e.g., '*_rpf.txt')."
    )
    input_group.add_argument(
        '-o', dest='output', required=False, type=str, help='output file name (default: RPF.file.list).'
    )

    args = parser.parse_args()
    args_dict = vars(args)
    for k, v in args_dict.items():
        print('{:<12}:  {:<}'.format(k, str(v)), flush=True)

    return args


def process_density_files(density_list):
    '''
    @Message  : merge the list file path and name.
    @Input    : list --> pattern for density files (e.g., '*_rpf.txt')
    @Return   : 
                output --> output dataframe contain the reads offset
    @Flow     : step1 --> get the file list
                step2 --> create the file list
    '''

    # create the output list
    density_merge = []

    # for each density file
    for density_file in density_list:
        file_name = os.path.basename(density_file)
        file_path_name = os.path.abspath(density_file)

        file_prefix = file_name.split('.')[0]
        file_prefix = file_prefix.replace('_rpf', '')
        file_prefix = file_prefix.replace('_rna', '')

        if file_name.endswith('_rpf.txt'):
            density_model = 'Ribo'
        elif file_name.endswith('_rna.txt'):
            density_model = 'RNA'
        else:
            print('Error: the file name is not correct, please check it.')
            exit()

        # merge file path and name

        density_merge.append([file_prefix, file_path_name, density_model])

    return density_merge, density_model


def output_table(density_merge, density_model, output_file):
    '''
    @Message  : function for output.
    @Input    : result_dict --> nested dict contain the reads offset
    @Return   : output --> description
    @Flow     : step1 --> convert the dict to dataframe
    '''

    # set the result name
    if output_file:
        output_file = output_file
    else:
        if density_model == 'Ribo':
            output_file = 'RPF.file.list'
        elif density_model == 'RNA':
            output_file = 'RNA.file.list'

    # output the result
    density_merge_df = pd.DataFrame(density_merge, columns=['Name', 'File', 'Type'])

    density_merge_df.to_csv(output_file, sep='\t', index=False)


def main():

    print('Step1: Checking the input Arguments.', flush=True)
    args = merge_density_args_parser()

    print('Step2: list the density file.', flush=True)
    density_file_list, density_model = process_density_files(args.list)

    print('Step3: output the density file list.', flush=True)
    output_table(density_file_list, density_model, args.output)

    print('All done.', flush=True)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project : riboParser
# @Script  : Pausing.py
# @Author  : Rensc

import sys
import matplotlib
import matplotlib.pyplot as plt
import seaborn as sns
import numpy as np
import pandas as pd
from . import RPFs


class Pausing(object):

    def __init__(self, args):
        # input and output files
        self.rpf = args.rpf
        self.output_prefix = args.output
        self.figure = args.figure
        self.all_pausing = args.all

        # opts for rpf density file parsering
        self.site = args.site
        self.frame = args.frame
        self.norm = args.normal

        self.tis = args.tis
        self.tts = args.tts
        self.rpf_num = args.min
        self.gene = args.list
        self.background = args.background

        self.sample_num = None
        self.sample_name = None

        self.raw_rpf = None
        self.merged_rpf = None

        self.gene_rpf_sum = None
        self.total_rpf_num = None
        self.high_gene = None
        self.high_rpf = None

        # opts for pausing
        self.scale = args.scale
        self.pausing = None
        self.merge_pausing = None
        self.merge_pausing_flt = None
        self.cds_pausing = None
        self.cds_codon_pausing = None
        self.codon_pausing = None

        # codon list
        self.stop = args.stop
        self.codon_dict, self.codon_table = RPFs.codon_table()
        
        if self.stop:
            del self.codon_dict['TAA']
            del self.codon_dict['TAG']
            del self.codon_dict['TGA']
            
            self.codon_table = self.codon_table[self.codon_table['Abbr'] != 'Stop']

    def read_rpf(self):
        '''
        1. import the rpf density
        2. convert the rpf density to rpm
        '''
        # raw_rpf, sample_name, sample_num, merged_rpf, total_rpf_num, gene_rpf_sum, high_gene, high_rpf
        rpf_results = RPFs.import_rpf(rpf_file=self.rpf,
                                      sites=self.site,
                                      frame=self.frame,
                                      sample_num=None,
                                      sample_name=None,
                                      tis = self.tis,
                                      tts = self.tts,
                                      gene=self.gene,
                                      rpf_num=self.rpf_num)
        # self.raw_rpf = rpf_results[0]
        # rpf_results[0] = ''
        self.sample_name = rpf_results[1]
        self.sample_num = rpf_results[2]
        self.merged_rpf = rpf_results[3]
        # rpf_results[3] = ''
        self.total_rpf_num = rpf_results[4]
        # self.gene_rpf_sum = rpf_results[5].copy()
        self.high_gene = rpf_results[6].copy()
        # self.high_rpf = rpf_results[7].copy()
        del rpf_results

        self.merged_rpf = self.merged_rpf[self.merged_rpf['codon'].isin(self.codon_dict.keys())]

        if self.norm:
            self.merged_rpf[self.sample_name] = self.merged_rpf[self.sample_name] * 1e6 / self.total_rpf_num
            # self.high_rpf = self.merged_rpf.loc[self.merged_rpf['name'].isin(self.high_gene)]
            # self.gene_rpf_sum = self.high_rpf.loc[:, ["name"] + self.sample_name].groupby("name")[self.sample_name].sum()

        self.high_rpf = self.merged_rpf.loc[self.merged_rpf['name'].isin(self.high_gene)]
        del self.merged_rpf

    def get_pausing_score(self):
        '''
        for each gene and codon
        pausing score means:
        1. use mean codon rpf density as the background
        (1). filter the high expression genes
        (2). pausing score = rpf density of each codon / mean of gene density
        (3). merge all gene pausing

        or

        2. use up/down-stream codon density as the background
        (1). filter the high expression genes
        (2). pausing score = codon density / mean of upstream and downstream [number > 0] codon density
        (3). merge all gene pausing

        '''

        high_rpf_group = self.high_rpf.groupby('name')

        if self.background == 0:
            merge_pausing_list = []
            # use the mean rpf for division
            now_num = 0

            for gene in high_rpf_group:
                # counter
                now_num += 1
                if now_num % 500 == 0:
                    print('Row: {rows}, {gene}.'.format(rows=now_num, gene=gene[0]), flush=True)                    

                # prepare the gene df
                cds_idx = gene[1]['region'] == 'cds'
                gene_pausing = gene[1].iloc[:, 0:6]
                gene_rpf = gene[1].loc[:, self.sample_name].copy()
                cds_rpf = gene_rpf[cds_idx].copy()

                # calculate the mean rpf
                if self.tts != 0:
                    denominator = cds_rpf.iloc[self.tis:-self.tts].mean()
                else:
                    denominator = cds_rpf.iloc[self.tis::].mean()

                # if 0 in denominator.values:
                #     continue

                # numerator = np.asarray(gene_rpf, dtype=float)

                # calculate the PS
                pausing_score = np.where(denominator == 0, 0, np.divide(gene_rpf, denominator))

                gene_pausing[self.sample_name] = pd.DataFrame(pausing_score).values
                # gene_pausing = pd.concat([gene_mess.iloc[:, 0:6], pd.DataFrame(pausing_score)], axis=1)
                merge_pausing_list.append(gene_pausing)

            self.merge_pausing = pd.concat(merge_pausing_list, axis=0, ignore_index=True)
            self.merge_pausing.replace([np.inf, -np.inf], np.nan)

            print('Row: {rows}, {gene}.'.format(rows=now_num, gene=gene[0]), flush=True)

        else:
            temp_rpf = pd.DataFrame(np.zeros((self.background * 2, self.sample_num)).astype(int))
            temp_rpf.columns = self.sample_name.copy()
            merge_pausing_list = []
            now_num = 0
            
            # calculate the ribosome pausing score with front and behind rpf
            for gene in high_rpf_group:
                now_num += 1
                if now_num % 500 == 0:
                    print('Row: {rows}, {gene}.'.format(rows=now_num, gene=gene[0]), flush=True)
                # if now_num == len(high_rpf_group):
                #     print('Row: {rows}, {gene}.'.format(rows=now_num, gene=gene[0]), flush=True)

                gene_rpf = gene[1][self.sample_name].copy()
                gene_pausing = gene[1].iloc[:, 0:6]

                # calculate the background mean rpf
                gene_rpf_rolling = pd.concat([gene_rpf, temp_rpf], ignore_index=True).rolling(self.background * 2 + 1).mean()
                # gene_rpf_rolling = temp_rpf.append([gene_rpf, temp_rpf],ignore_index=True).rolling(self.background * 2 + 1).mean()
                gene_rpf_rolling = gene_rpf_rolling.loc[self.background * 2::]

                # numerator = np.asarray(gene_rpf, dtype=float)
                denominator = np.asarray(gene_rpf_rolling, dtype=float)
                pausing_score = np.where(denominator == 0, 0, np.divide(gene_rpf, denominator))

                gene_pausing[self.sample_name] = pd.DataFrame(pausing_score).values

                # gene_pausing = pd.concat([gene_mess.iloc[:, 0:6], pd.DataFrame(pausing_score)], axis=1)
                merge_pausing_list.append(gene_pausing)
            
            print('Row: {rows}, {gene}.'.format(rows=now_num, gene=gene[0]), flush=True)

            self.merge_pausing = pd.concat(merge_pausing_list, axis=0, ignore_index=True)
            self.merge_pausing.replace([np.inf, -np.inf], np.nan)
        del high_rpf_group

    def output_cds_pausing(self):
        # cds = self.merge_pausing[self.merge_pausing['region'] == 'cds']
        cds_group = self.merge_pausing.groupby('name')
        cds_length = cds_group[['name', 'from_tis']].tail(1).set_index('name', drop=True)
        cds_length.columns = ['aa_num']
        cds_pausing = cds_group[self.sample_name].sum()
        self.cds_pausing = pd.concat([cds_length, cds_pausing], axis=1)

        out_gene_txt = self.output_prefix + "_cds_pausing_score.txt"
        self.cds_pausing.to_csv(out_gene_txt, sep='\t', index=True)

        # utr5_pausing = self.merge_pausing[self.merge_pausing['region'] == '5utr']
        # gene_utr5_pausing = utr5_pausing.groupby('name')[self.sample_name].sum()
        # utr3_pausing = self.merge_pausing[self.merge_pausing['region'] == '3utr']
        # gene_utr3_pausing = utr3_pausing.groupby('name')[self.sample_name].sum()

    def output_cds_codon_pausing(self):
        out_cds_codon_txt = self.output_prefix + "_cds_codon_pausing_score.txt"
        titles = ['Name', 'Codon'] + [i + '_sum' for i in self.sample_name] + [i + '_mean' for i in self.sample_name]
        titles = pd.DataFrame(titles).T
        titles.to_csv(out_cds_codon_txt, sep='\t', index=False, header=False)

        # cds = self.merge_pausing[self.merge_pausing['region'] == 'cds']
        self.cds_codon_pausing = pd.pivot_table(self.merge_pausing, index=['name', 'codon'],
                                                values=self.sample_name, aggfunc=[np.sum, np.mean])
        self.cds_codon_pausing.to_csv(out_cds_codon_txt, sep='\t', mode='a', header=False)


    @staticmethod
    def scale_method(scale, pausing_score):
        if scale == 'minmax':
            # relative_pausing_score = (pausing_score - pausing_score.min()) / (pausing_score.max() - pausing_score.min())
            relative_pausing_score = pausing_score / pausing_score.max()
        elif scale == 'zscore':
            relative_pausing_score = (pausing_score - pausing_score.mean()) / pausing_score.std()
        else:
            print('Unknown scale method.', flush=True)
            sys.exit()
        return relative_pausing_score


    def output_sum_codon_pausing(self):
        '''
        @Message  : output the sum of codon pausing score
        @Input    : self.high_rpf --> high express gene rpf
        @Return   : self.codon_pausing --> codon pausing score
        @Flow     : step1 --> summary the total and valid codon number
                    step2 --> summary the pausing score
                    step3 --> calculate the relative pausing score
                    step4 --> merge pausing score of total and valid codon
        '''

        # self.codon_table = self.codon_table.reset_index(drop=False).sort_values('codon')

        # sum the valid codon num and count
        cds_rpf = self.high_rpf[self.high_rpf['region'] == 'cds']
        total_codon = cds_rpf['codon'].value_counts()

        val_rpf = cds_rpf[self.sample_name] > 0
        val_rpf.loc[:, 'codon'] = cds_rpf['codon'].copy()
        valid_codon = val_rpf.groupby('codon')[self.sample_name].sum()
        rpf_count = cds_rpf.groupby('codon')[self.sample_name].sum()

        # rename or format
        total_codon.columns = 'total_codon'
        valid_codon.columns = [i + '_valid_codon' for i in self.sample_name]
        rpf_count.columns = [i + '_rpf_count' for i in self.sample_name]

        # summary the total codon pausing score
        total_codon_pausing = self.merge_pausing.groupby('codon')[self.sample_name].mean()
        total_codon_pausing.columns = [i + '_absolute_total_ps' for i in self.sample_name]

        valid_pausing = self.merge_pausing[self.merge_pausing[self.sample_name].max(axis=1) > 0]
        valid_codon_pausing = valid_pausing.groupby('codon')[self.sample_name].mean()
        valid_codon_pausing.columns = [i + '_absolute_valid_ps' for i in self.sample_name]

        # calculate the relative pausing score
        relative_total_codon_pausing = self.scale_method(self.scale, total_codon_pausing)
        relative_valid_codon_pausing = self.scale_method(self.scale, valid_codon_pausing)

        relative_total_codon_pausing.columns = [i + '_relative_total_ps' for i in self.sample_name]
        relative_valid_codon_pausing.columns = [i + '_relative_valid_ps' for i in self.sample_name]

        # ouput the total codon pausing score
        out_codon_txt = self.output_prefix + "_sum_codon_pausing_score.txt"
        self.codon_pausing = pd.concat([self.codon_table, total_codon, valid_codon, rpf_count, total_codon_pausing,
                                       valid_codon_pausing, relative_total_codon_pausing, relative_valid_codon_pausing], 
                                       axis=1, join='inner')
        
        self.codon_pausing.index.name = 'Codon'
        self.codon_pausing = self.codon_pausing.sort_values('Abbr').reset_index(drop=False)
        self.codon_pausing.sort_values(by=['Abbr', 'Codon'], ascending=True, inplace=True)
        self.codon_pausing.to_csv(out_codon_txt, sep='\t', index=False)

    def output_all_pausing(self):
        if self.all_pausing:
            all_pausing_txt = self.output_prefix + "_all_pausing_score.txt"
            self.merge_pausing.to_csv(all_pausing_txt, sep='\t', index=False)

    def draw_codon_total_pausing_heat(self):
        '''
        @Message  : draw the codon total pausing score heatmap
        @Input    : self.codon_pausing --> codon pausing score
                    self.output_prefix --> prefix of output filename
        @Return   : output --> codon total pausing score heatmap
        @Flow     : step1 --> select the column of total pausing score
                    step2 --> draw the heatmap
        '''
        
        out_pdf = "{prefix}_total_pausing_heatplot.pdf".format(prefix=self.output_prefix)
        out_png = "{prefix}_total_pausing_heatplot.png".format(prefix=self.output_prefix)

        ps_title = [i + '_absolute_total_ps' for i in self.sample_name]

        # select the column of total pausing score
        total_codon_pausing = self.codon_pausing[['Codon', 'Abbr'] + ps_title].copy()

        # draw the rpf pausing score
        matplotlib.use('AGG')
        fig = plt.figure(figsize=(10 + 0.5 * self.sample_num, 20), dpi=300)
        sns.heatmap(data=total_codon_pausing[ps_title].astype('float16'),
                    yticklabels=total_codon_pausing['Codon'] + '[' + total_codon_pausing['Abbr'] + ']',
                    annot=True, fmt=".3f", cmap="vlag", linewidths=.01)

        # plt.suptitle("codon pausing")
        plt.tight_layout()
        # plt.show()

        fig.savefig(fname=out_pdf)
        fig.savefig(fname=out_png)

    def draw_codon_valid_pausing_heat(self):
        '''
        @Message  : draw the codon total pausing score heatmap
        @Input    : self.codon_pausing --> codon pausing score
                    self.output_prefix --> prefix of output filename
        @Return   : output --> codon valid pausing score heatmap
        @Flow     : step1 --> select the column of valid pausing score
                    step2 --> draw the heatmap
        '''

        out_pdf = "{prefix}_valid_pausing_heatplot.pdf".format(prefix=self.output_prefix)
        out_png = "{prefix}_valid_pausing_heatplot.png".format(prefix=self.output_prefix)

        ps_title = [i + '_absolute_valid_ps' for i in self.sample_name]

        # select the column of valid pausing score
        valid_codon_pausing = self.codon_pausing[['Codon', 'Abbr'] + ps_title].copy()

        # draw the rpf pausing score
        matplotlib.use('AGG')
        fig = plt.figure(figsize=(10 + 0.5 * self.sample_num, 20), dpi=300)
        sns.heatmap(data=valid_codon_pausing[ps_title].astype('float16'),
                    yticklabels=valid_codon_pausing['Codon'] + '[' + valid_codon_pausing['Abbr'] + ']',
                    annot=True, fmt=".3f", cmap="vlag", linewidths=.01)

        # plt.suptitle("codon pausing")
        plt.tight_layout()
        # plt.show()

        fig.savefig(fname=out_pdf)
        fig.savefig(fname=out_png)

    def draw_codon_pausing_plot(self, category):
        out_pdf = "{prefix}_{category}_pausing_plot.pdf".format(prefix=self.output_prefix, category=category)
        out_png = "{prefix}_{category}_pausing_plot.png".format(prefix=self.output_prefix, category=category)

        # draw the rpf pausing score
        matplotlib.use('AGG')
        fig = plt.figure(figsize=(9, 3 * self.sample_num), dpi=300)

        for fig_num in range(self.sample_num):
            ax = fig.add_subplot(self.sample_num, 1, fig_num + 1)
            sns.barplot(x=self.codon_pausing.index,
                        y=self.codon_pausing[self.sample_name[fig_num] + '_' + category + '_ps'],
                        ax=ax)
            ax.set_xticks(self.codon_pausing.index)
            ax.set_xticklabels(self.codon_pausing['Codon'] + " [" + self.codon_pausing['Abbr'] + "]",
                               size=8,
                               va='top')
            ax.set_xlim([-1, 65])

            # add the gene annotation here, need to import the txt file
            ax.set_ylabel('Pausing score')
            ax.set_xlabel('Codon')
            ax.set_title(self.sample_name[fig_num])
            plt.xticks(rotation=90)

        plt.suptitle("{category} codon pausing".format(category=category))
        plt.tight_layout()
        # plt.show()

        fig.savefig(fname=out_pdf)
        fig.savefig(fname=out_png)

    @staticmethod
    def get_labels(now_gene):
        start_site = now_gene["from_tis"].iloc[0]
        start_codon_site = 0
        stop_codon_site = int(now_gene[now_gene["from_tts"] == 0]["from_tis"])
        stop_site = now_gene["from_tis"].iloc[-1]

        now_labels = [start_site, start_codon_site, stop_codon_site, stop_site]

        return now_labels

    def draw_rpf_pausing_plot(self):

        merge_pausing_group = self.merge_pausing.groupby('name')

        for gene in merge_pausing_group:
            print("Now gene: {gene}.".format(gene=gene[0]), flush=True)

            now_gene_pausing = gene[1]
            now_labels = self.get_labels(now_gene_pausing)

            # draw the rpf pausing score
            matplotlib.use('AGG')
            fig = plt.figure(figsize=(9, 3 * self.sample_num), dpi=300)

            for fig_num in range(self.sample_num):
                ax1 = fig.add_subplot(self.sample_num, 1, fig_num + 1)
                ax1.bar(now_gene_pausing.from_tis,
                        now_gene_pausing[self.sample_name[fig_num]],
                        width=2,
                        color="#f86934")

                ax1.set_xticks(now_labels)
                ax1.set_xticklabels(now_labels)
                ax1.set_xlim([now_labels[0], now_labels[-1]])

                # add the gene annotation here, need to import the txt file
                ax1.set_ylabel('Pausing score')
                ax1.set_xlabel('Position (nt)')
                ax1.set_title(self.sample_name[fig_num])
                plt.xticks(rotation=90)

            plt.suptitle(str(gene[0]))
            plt.tight_layout()
            # plt.show()

            if self.figure == 'pdf':
                out_pdf = str(gene[0]) + "_pausing_plot.pdf"
                fig.savefig(fname=out_pdf)
            elif self.figure == 'png':
                out_png = str(gene[0]) + "_pausing_plot.png"
                fig.savefig(fname=out_png)
            else:
                print('Unknown figure format.', flush=True)
                sys.exit()
            plt.close()
